{----------------------------------------------------------------------------
|
| Library: Spider Object Database
|
| Module: Odb.Pas
|
| Description: Core classes.
|
| History: Dec 29, 1997. Michel Brazeau, first version
|          Jan 28, 1998. MB Merged OdbDrv.Pas, OdbFile.Pas, and OdbDataB units
|                        into Odb.Pas to prevent circular unit references.
|          Apr 11, 1998. MB removed unnecessary parameters to
|                        NewHandle, renamed GetObjectFileData to
|                        GetHandleDataObj and return data object directly rather than
|                        the TObjectFileData object.
|          May 13, 1998. MB add ObjectFile parameter to InternalDeleteFile
|          Sep 07, 2000. MB start architectural changes, with new TObjectDataset
|                        base class.
|          Feb 16, 2001. MB TOdbDriver now has introspection properties to determine
|                        it capabilities
|          Mar 14, 2001. MB Add TORMapping class. Transfer DefineField,
|                        DefineIndices, etc..., responsibilities to
|                        TORMapping. Once an TORMapping is created, it may be
|                        shared among TObjectFile, TObjectQuery, TObjectUpdate,
|                        and there is not need to derive classes from TObjectFile
|                        or TObjectQuery
|          Mar 16, 2001. MB Remove event handlers, as there are virtual methods
|                        that may be overriden. These methods are now all
|                        mostly handled in the TORMapping class.
|
|---------------------------------------------------------------------------}
unit Odb;

{$I Odb.Inc}

interface

uses
    SysUtils, { for Exception }
    Classes,  { for TMemoryStream }
    ObjArray, { for TArray }
    ObjList,  { for TUnOrderedList }
    OdbIndex, { for TOdbIndex }
    OdbField; { for TOdbFieldType }

type

{ Handle type used by ODB objects. }
TOdbHandle  = LongInt;

TOdbDriver       = class;
TObjectDatabase  = class;
TObjectDataset   = class;
TObjectFile      = class;

TORMapping       = class;
TORMappingClass  = class of TORMapping;

{--------------------------------------------------------------------------}

{ Bookmark class used by GetBookmark, GotoBookmark methods of TObjectFile. }
TOdbBookmark = class(TObject)
protected
    FIndexName     : String;
    FObjectFile    : TObjectFile;
    FBookmarkData  : Pointer;

public
    constructor Create( const ObjectFile : TObjectFile );
    destructor Destroy; override;

    property IndexName : String read FIndexName;
end;

{ Abstract class to model a set of objects. This serves as the base class for
  object files (analogous to a table in a relational database), and the results
  of a query
}
TObjectDataset = class(TObject)
protected
    FHandle         : TOdbHandle;
    FCurrentObj     : TObject;
    FDatabase       : TObjectDatabase;
    FORMapping      : TORMapping;

    { the DatasetName is the name of the table in the database }
    FDatasetName    : String;

    function GetDriver : TOdbDriver;

    function GetCurrentObj : TObject;

    function GetOpened : Boolean;

    function GetObjectType : TClass;
    function GetFields : TUnOrderedList;
    function GetIndices : TUnOrderedList;

public
    { Create the dataset associated to InDatabase. }
    constructor Create( const InDatabase        : TObjectDatabase;
                        const InDatasetName     : String;
                        const InORMappingClass  : TORMappingClass );

    destructor Destroy; override;

    property ObjectType : TClass read GetObjectType;

    property ORMapping : TORMapping read FORMapping;

    { Current object positioned. If it is assigned a new object, the current
      object is freed. }
    property CurrentObj : TObject read GetCurrentObj;

    property Database : TObjectDatabase read FDatabase;

    property DatasetName : String read FDatasetName;

    { list of fields participating on indices. }
    property Fields : TUnOrderedList read GetFields;

    { returns a field given a field name. May raise EOdbFieldDoesNotExist }
    function FieldByName( const Name : String ) : TOdbField;

    { list indexes. }
    property Indices : TUnOrderedList read GetIndices;

    property Driver : TOdbDriver read GetDriver;

    { handle used with TOdbDriver instance. }
    property Handle : TOdbHandle read FHandle;

    { physically create the dataset in the database. }
    procedure CreateDataset;

    { physically delete the dataset from the database }
    procedure DeleteDataset;

    { open the data set }
    procedure Open;

    { close the data set }
    procedure Close;

    { returns True if file is opened }
    property Opened : Boolean read GetOpened;

    { reloads the current object positioned into CurrentObj. }
    procedure ReloadCurrent;

    { returns True if the data set is empty. }
    function IsEmpty : Boolean;

    { set the current position to the next object. Returns False if at EOF. }
    function FindNext : Boolean;

    { set the current position to the previous object. Returns False if at
      BOF. }
    function FindPrior : Boolean;

    { set the current position to the first object. Returns False if file is
      empty. }
    function FindFirst : Boolean;

    { set the current position to the last object. Returns False if file is
      empty. }
    function FindLast : Boolean;


protected


    { raises EInvalidObjectType if Obj is not if type FObjectType. }
    procedure CheckObjectType( const Obj : TObject );

    { returns a list of all unique TOdbIndex references. }
    procedure GetAllUniqueIndices( const IndexList : TUnOrderedList );

    { returns a list of all unique TOdbIndex references which are modified
      given previous and new field values. }
    procedure GetModifiedUniqueIndices( const IndexList      : TUnOrderedList;
                                        const PreviousFields : TUnOrderedList;
                                        const CurrentFields  : TUnOrderedList );

    procedure BeforeInsert; virtual;
    procedure AfterInsert; virtual;

    procedure BeforeUpdate; virtual;
    procedure AfterUpdate; virtual;

    procedure BeforeDelete; virtual;
    procedure AfterDelete; virtual;

    procedure BeforeObjectPost( const Stream  : TMemoryStream ); virtual;
    procedure AfterObjectRead( const Stream : TMemoryStream ); virtual;

    procedure BeforeOpen; virtual;
    procedure AfterOpen; virtual;

    procedure BeforeClose; virtual;
    procedure AfterClose; virtual;


end;

{--------------------------------------------------------------------------}

{ Object file class. An object file is a file containing objects, analogous
to a table in a relational database.

The following methods must be overriden:

  UpdateFieldsFromObj
  DefineIndices
  DefineFields

If an autoincrement field is used, UpdateObjFromFields must be overriden.

BeforeObjectPost and AfterObjectRead may be overriden to provide
compression and/or encrytion.
}
TObjectFile = class(TObjectDataset)
protected

    function GetObjectCount : LongInt;

    function GetIndexName : String;
    procedure SetIndexName( const InIndexName : String );

    procedure DefineFields; virtual;
    procedure DefineIndices; virtual;

    function GetCurrentIndex : TOdbIndex;

    procedure SetCurrentObj( const Obj : TObject );

public
    constructor Create( const InDatabase       : TObjectDatabase;
                        const InDatasetName    : String;
                        const InORMappingClass : TORMappingClass );

    { Constructor for backward compatibility with version 1.X of Spider ODB.

      The following methods must be overriden, when using this compatiblity mode:
        UpdateFieldsFromObj
        DefineFields
        DefineIndices
        UpdateObjFromFields, if autoincrement field is used }
    constructor CreateVersion1( const InDatabase       : TObjectDatabase;
                                const InObjectFileName : String;
                                const InObjectType     : TClass );

    { CurrentObj with a TObjectFile is writable }
    property CurrentObj : TObject read GetCurrentObj
                                  write SetCurrentObj;

    { insert CurrentObj in the data set }
    procedure Insert;

    { delete the current object positioned from the data set. The next
      object then becomes the new current object. If the object deleted was
      the last object in the file, then the previous object becomes the
      current object. }
    procedure Delete;

    { update CurrentObj }
    procedure Update;

    property ObjectCount : LongInt read GetObjectCount;

    { ObjectFileName is added for backward compatibility with Spider ODB
      version 1 }
    property ObjectFileName : String read FDatasetName;

    { current index. FindXXX methods are sensitive to the current
      index. }
    property IndexName : String read GetIndexName
                                write SetIndexName;

    { MB Sep 08, 2000. CreateFile and DeleteFile are added for backward
      compatibility with Spider ODB version 1. They simply call CreateDataset
      and DeleteDataset respectively }
    procedure CreateFile;
    procedure DeleteFile;

    { returns a TOdbIndex reference corresponding to the current index
      identified by IndexName }
    property CurrentIndex : TOdbIndex read GetCurrentIndex;

    { Update the field values in Fields from the contents of Obj. }
    procedure UpdateFieldsFromObj( const Obj    : TObject ); virtual;

    { Update Obj from the values in Fields. This method should only be
      overriden when the object file is defined with an autoincrement field. }
    procedure UpdateObjFromFields( const Obj    : TObject ); virtual;

    { open the object file with exclusive access. }
    procedure OpenExclusive;

    { lock the current object positioned. Returns True if lock acquired
      successfully. }
    function LockCurrent : Boolean;

    { unlock the current object positioned. }
    procedure UnLockCurrent;

    { find the object which is equal to the data set in CurrentObj. }
    function FindExact : Boolean;

    { find the object which is equal to or greater than the data set in
      CurrentObj. If all objects are smaller, the last object is positioned. }
    procedure FindNearest;

    { returns a bookmark on the current position. After usage Bookmark must be
      freed explicitly, or freed by the GotoBookmark method. }
    function GetBookmark : TOdbBookmark;

    { restores the position saved in a bookmark. If FreeBookmark is True the
      bookmark will be freed after the position is restored. }
    procedure GotoBookmark( const Bookmark     : TOdbBookmark;
                            const FreeBookmark : Boolean );

    { pack the table (remove "soft" deleted records). Some drivers will
      ignore this request when not applicable. Most drivers require exclusive
      access to the file. }
    procedure Pack;

    { raises EOdbKeyViolation if under any of the indices in InIndexList there
      exists an object which matches the field values in InFields. }
    procedure CheckForDuplicate( const InIndexList : TUnOrderedList;
                                 const InFields    : TUnOrderedList );

    { MB Feb 16, 2001. Made AddField and AddIndex public. See
      .\Tests\TestObj.Pas }

    { AddField method should be used inside the derived DefineFields
      method. The field added is returned. Obsolete in version 2.
      See TORMapping }
    function AddField( const InName       : String;
                       const InType       : TOdbFieldType;
                       const InSize       : Byte ) : TOdbField;

    { AddIndex method should be used inside the derived DefineIndices
      method. The index just added is returned.  Obsolete in version 2.
      See TORMapping}
    function AddIndex( const IndexName  : String;
                       const Attributes : TOdbIndexAttributes;
                       const FieldNames : array of String ) : TOdbIndex;

end;

{--------------------------------------------------------------------------}

{ Object query class. An object query is a set of objects defined by an
  SQL query. The query is specified with respect to the fields defined
  in the main dataset, but may include fields in other datasets part of the
  database. The object itself is always retrieved from the main dataset.

  The TObjectQuery is a "query" that is, it returns information but does not
  modify it. To update objects, use a TObjectUpdate instance }

TObjectQuery = class(TObjectDataset)
protected
    FSelect : TStringList;

public

    constructor Create( const InDatabase       : TObjectDatabase;
                        const InDatasetName    : String;
                        const InORMappingClass : TORMappingClass );

    destructor Destroy; override;

    { This is the select statement to populate the dataset. }
    property Select : TStringList read FSelect;

    { simple query which attempts to locate an object, given an object from
      which the key will be obtained according to the primary key of the
      dataset. Returns the object found or nil if Key is not found. The returned
      object must NOT be destroyed.
      If the query is currently opened,
      it will first be closed. On return the query is closed. }
    function Search( const Key      : TObject ) : TObject;
end;

{--------------------------------------------------------------------------}

TObjectUpdate = class(TObject)
protected
    FHandle         : TOdbHandle;
    FDatabase       : TObjectDatabase;
    FORMapping      : TORMapping;

    FDatasetName    : String;

    FBeforeInsert   : TNotifyEvent;
    FAfterInsert    : TNotifyEvent;
    FBeforeUpdate   : TNotifyEvent;
    FAfterUpdate    : TNotifyEvent;
    FBeforeDelete   : TNotifyEvent;
    FAfterDelete    : TNotifyEvent;
    FBeforePost     : TNotifyEvent;
    FAfterRead      : TNotifyEvent;
public
    constructor Create( const InDatabase       : TObjectDatabase;
                        const InDatasetName    : String;
                        const InORMappingClass : TORMappingClass );

    destructor Destroy; override;

    property Database : TObjectDatabase read FDatabase;

    { Insert Obj in the database table defined by the dataset.  }
    procedure Insert( const Obj : TObject );

    { If an object is found in the related database table that has the same
      key value as defined by the primary index of the table, the object is
      deleted from the table. }
    procedure Delete( const Obj : TObject );

    { If an object is found in the related database table that has the same
      key value as defined by the primary index of the table, the object is
      updated in the table.  }
    procedure Update( const Obj : TObject );

end;

{--------------------------------------------------------------------------}

{ Class which defines an object to relational mapping. This allows sharing the
  same mapping code between TObjectFile, TObjectQuery, and TObjectUpdate }
TORMapping = class(TObject)
protected
    FFields         : TUnOrderedList;   { of TOdbField }
    FIndices        : TUnOrderedList;   { of TOdbIndex }
    FDatabase       : TObjectDatabase;

    function GetObjectType : TClass; virtual; abstract;

    function AddField( const InName       : String;
                       const InType       : TOdbFieldType;
                       const InSize       : Byte ) : TOdbField;

    function AddIndex( const IndexName  : String;
                       const Attributes : TOdbIndexAttributes;
                       const FieldNames : array of String ) : TOdbIndex;

    procedure DefineFields; virtual; abstract;
    procedure DefineIndices; virtual; abstract;

    function FieldByName( const Name : String ) : TOdbField;

public
    constructor Create( const InDatabase : TObjectDatabase ); virtual;
    destructor Destroy; override;

    property ObjectType : TClass read GetObjectType;

    property Fields : TUnOrderedList read FFields;
    property Indices : TUnOrderedList read FIndices;

    { Returns the primary index defined for the dataset. Raises
      EOdbTableMustHavePrimaryIndex if a primary index is not defined. The
      returned object must NOT be destroyed }
    function GetPrimaryIndex( const DatasetName : String ) : TOdbIndex;

    procedure UpdateFieldsFromObj( const Obj       : TObject ); virtual; abstract;

    procedure UpdateObjFromFields( const Obj       : TObject ); virtual;

    procedure BeforeInsert; virtual;
    procedure AfterInsert; virtual;

    procedure BeforeUpdate; virtual;
    procedure AfterUpdate; virtual;

    procedure BeforeDelete; virtual;
    procedure AfterDelete; virtual;

    procedure BeforeObjectPost( const Stream  : TMemoryStream ); virtual;
    procedure AfterObjectRead( const Stream : TMemoryStream ); virtual;

end;

{--------------------------------------------------------------------------}

{ Class used internally be a TOdbDriver instance which relates a general
purpose data object with a handle of a dataset of an object update. For instance,
the BDE driver uses stores a TTable instance in DataObj.}
THandleData = class(TObject)
protected
    FHandle        : LongInt;
    FDataObj       : TObject;

public
    { Object file handle. }
    property Handle : LongInt read FHandle write FHandle;

    { Data object related to object dataset. }
    property DataObj : TObject read FDataObj write FDataObj;

    constructor Create( const InHandle        : LongInt;
                        const InDataObj       : TObject ); virtual;

    destructor Destroy; override;
end;

{--------------------------------------------------------------------------}

TDataAccessMethod = (
                      damISAM,
                      damSQL
                     );

TDataAccessMethodSet = set of TDataAccessMethod;

{ abstract object database driver

  The following methods MUST be implemented by a derived class:

    InternalOpenDataset
    InternalCloseDataset

    InternalCreateDataset
    InternalDeleteDataset

    InternalGetCurrent

    GetSupportedFieldTypes

    GetDataAccessMethods

    InternalFindNext
    InternalFindPrior
    InternalFindFirst
    InternalFindLast



    ISAM driver only:
    InternalInsert
    InternalUpdate
    InternalDelete
    InternalFindExact
    InternalFindNearest
    InternalGetBookmark
    InternalGotoBookmark
    InternalFreeBookmark
    InternalGetIndexName
    InternalSetIndexName
    InternalGetObjectCount

    If the driver supports "soft" deleted records and "packing",
    InternalPack

    SQL driver only:
    InternalInsertSQL
    InternalUpdateSQL
    InternalDeleteSQL

    InternalCreateObjectUpdate
    InternalDestroyObjectUpdate
}
TOdbDriver = class(TObject)
protected
    { Array of THandleData instances for all opened datasets in the
    database. }
    FHandles               : TArray;

    FDatabaseName          : String;
    FDatabaseOpened        : Boolean;

    FSupportedFieldTypes   : TOdbFieldTypes;

    { Checks if the driver supports the specified access method. If the driver
      does not support it, the EOdbUnsupportedAccessMethod exception is
      raised. }
    procedure CheckAccessMethod( const AccessMethod : TDataAccessMethod );

public
    { Creates an object database driver to access the database identified by
      DatabaseName. }
    constructor Create(const InDatabaseName : String); virtual;
    destructor Destroy; override;

    { Name of the database. For BDE or server based databases, the database
      name in the name registered with the server. For desktop databases, the
      database name is a directory. }
    property DatabaseName : String read FDatabaseName;

    { Start a transaction on the database. Commit must later be invoked to
    accepted the modifications made to the database, or Rollback to cancel the
    modifications made to the database. }
    procedure StartTransaction;

    { Accept the modifications since the previous call to StartTransaction.
    Most databases do not support nested transactions. }
    procedure Commit;

    { Cancel the modications since the previous call to StartTransaction. }
    procedure Rollback;

    { Open a session with the database. For desktop databases, the parameters
    may be blank. }
    procedure OpenDatabase( const ServerName : String;
                            const UserName   : String;
                            const Password   : String
                          );

    { Close the current database session. }
    procedure CloseDatabase;

    { Returns True if the database is currently opened. }
    property DatabaseOpened : Boolean read FDatabaseOpened;

    { Open the database file identified by Name. ObjectFile may be used by the
    driver to ensure the properties (fields, indices) of the physical file match
    the properties defined in the definition of the object file. The driver
    must return a handle to the file. }
    function OpenDataset( const ObjectDataset : TObjectDataset ) : TOdbHandle;

    { Same as OpenDataset, but opens the dataset with exclusive access. }
    function OpenExclusive( const ObjectFile : TObjectFile ) : TOdbHandle;

    { Close the file associated with Handle. }
    procedure CloseDataset( const Handle : TOdbHandle );

    { Create a physical file having the properties defined in ObjectFile. }
    procedure CreateDataset( const ObjectDataset : TObjectDataset );

    procedure DeleteDataset( const ObjectDataset : TObjectDataset );

    procedure Insert( const Handle      : TOdbHandle;
                      const Fields      : TUnOrderedList;
                      const Stream      : TMemoryStream);
    procedure Update( const Handle     : TOdbHandle;
                      const Fields     : TUnOrderedList;
                      const Stream     : TMemoryStream );
    procedure Delete( const Handle : TOdbHandle );

    procedure InsertSQL( const Handle       : TOdbHandle;
                         const Stream       : TMemoryStream;
                         const ORMapping    : TORMapping;
                         const DatasetName  : String );

    procedure UpdateSQL( const Handle      : TOdbHandle;
                         const Stream      : TMemoryStream;
                         const ORMapping   : TORMapping;
                         const DatasetName : String );

    procedure DeleteSQL( const Handle      : TOdbHandle;
                         const ORMapping   : TORMapping;
                         const DatasetName : String );


    function GetIndexName( const Handle : TOdbHandle ) : String;
    procedure SetIndexName( const Handle    : TOdbHandle;
                            const IndexName : String );

    procedure GetCurrent( const Handle  : TOdbHandle;
                          const Stream  : TMemoryStream );

    function LockCurrent( const Handle : TOdbHandle ) : Boolean;
    procedure UnLockCurrent( const Handle : TOdbHandle );

    function FindNext( const Handle : TOdbHandle ) : Boolean;
    function FindPrior( const Handle : TOdbHandle ) : Boolean;
    function FindFirst( const Handle : TOdbHandle ) : Boolean;
    function FindLast( const Handle : TOdbHandle ) : Boolean;
    function FindExact( const Handle : TOdbHandle;
                        const Fields : TUnOrderedList;
                        const Index  : TOdbIndex ) : Boolean;
    procedure FindNearest( const Handle : TOdbHandle;
                           const Fields : TUnOrderedList;
                           const Index  : TOdbIndex );

    function GetBookmark( const Handle : TOdbHandle ) : Pointer;
    procedure GotoBookmark( const Handle   : TOdbHandle;
                            const Bookmark : Pointer );
    procedure FreeBookmark( const Handle   : TOdbHandle;
                            const Bookmark : Pointer );

    { Returns the number of objects in the file. For DBase tables this may
      not be accurate as "soft" deleted records are still counted. }
    function GetObjectCount( const Handle : TOdbHandle ) : LongInt;

    { Returns True if the file does not contain any objects. The default
      implementation uses GetObjectCount=0, but for Dbase tables this
      expression is not accurate. }
    function IsEmpty( const Handle : TOdbHandle ) : Boolean;

    procedure Pack( const Handle : TOdbHandle );

    property SupportedFieldTypes : TOdbFieldTypes read FSupportedFieldTypes;

    { override these routines to return True if the TObjectFile instance
      should check for duplicates and raise EOdbDuplicateKey on Insert and
      Update operations respectively. Some drivers may not support or handle
      properly unique keys. Ie, BDE with DBase tables does not raise a key
      violations exception even with the unique index attributes. }
    function CheckDuplicateInsert : Boolean; virtual;
    function CheckDuplicateUpdate : Boolean; virtual;

    { returns the data access methods supported by the driver. By default,
      [damISAM] is returned for compatibility with Spider ODB version
      1 drivers }
    function GetDataAccessMethods : TDataAccessMethodSet; virtual;

    { These methods must be overriden }
    function AutoIncSupported              : Boolean; virtual; abstract;
    function IntegerIndexSupported         : Boolean; virtual; abstract;
    function FloatIndexSupported           : Boolean; virtual; abstract;
    function CaseInsensitiveIndexSupported : Boolean; virtual; abstract;
    function DescendingIndexSupported      : Boolean; virtual; abstract;
    function TransactionsSupported         : Boolean; virtual; abstract;
    function FileLockingSupported          : Boolean; virtual; abstract;
    function ObjectLockingSupported        : Boolean; virtual; abstract;
    function SingleIndexOnly               : Boolean; virtual; abstract;

protected

    { check if the database is currently opened. May raise
      EOdbDatabaseNotOpened. }
    procedure CheckDatabaseOpened;

    { returns an allocated handle.  }
    function NewHandle( const InDataObj       : TObject ) : TOdbHandle;

    { release a handle and free its associated data. }
    procedure ReleaseHandle( const Handle : TOdbHandle );

    { returns the data object associated with a dataset or object update
      handle. }
    function GetHandleDataObj( const Handle : TOdbHandle ) : TObject;

    procedure InternalStartTransaction; virtual;
    procedure InternalCommit; virtual;
    procedure InternalRollback; virtual;

    procedure InternalOpenDatabase(
                            const ServerName : String;
                            const UserName   : String;
                            const Password   : String
                          ); virtual;
    procedure InternalCloseDatabase; virtual;

    { object file related methods. }

    function InternalOpenDataset( const ObjectDataset : TObjectDataset ) : TOdbHandle; virtual; abstract;

    function InternalOpenExclusive( const ObjectFile : TObjectFile ) : TOdbHandle; virtual;

    procedure InternalCloseDataset( const Handle : TOdbHandle ); virtual; abstract;

    function InternalCreateObjectUpdate : TOdbHandle; virtual; abstract;
    procedure InternalDestroyObjectUpdate( const Handle : TOdbHandle ); virtual; abstract;

    procedure InternalDeleteDataset( const ObjectDataset : TObjectDataset ); virtual; abstract;

    { create a dataset identified by Name. Fields is a list of TOdbField's to be
      indexed, and Indices is a list of TOdbIndex's for the file. }
    procedure InternalCreateDataset( const ObjectDataset : TObjectDataset ); virtual; abstract;

    { Insert an object in a dataset with the contents of Stream. If the object
      file is defined with an autoincrement field, the value of the
      autoincrement field must be updated in Fields. This method need only
      be overriden for a driver that supports the damISAM data
      access method. }
    procedure InternalInsert(
                      const Handle  : TOdbHandle;
                      const Fields  : TUnOrderedList;
                      const Stream  : TMemoryStream ); virtual; abstract;

    { update the current object in a dataset with the contents of Stream.
      This method need only be overriden for a driver that supports the
      damISAM data access method. }
    procedure InternalUpdate(
                      const Handle     : TOdbHandle;
                      const Fields     : TUnOrderedList;
                      const Stream     : TMemoryStream ); virtual; abstract;

    { delete the current object in a dataset. Positioning must be established
      to the next object if possible, if not to the previous. This method need
      only be overriden for a driver that supports the damISAM data
      access method. }
    procedure InternalDelete( const Handle : TOdbHandle ); virtual; abstract;

    { Insert an object in the dataset with the contents of Stream. This method
      need only be overriden for a driver that supports the damSQL data
      access method. Handle is related to a TObjectUpdate. }
    procedure InternalInsertSQL(
                      const Handle      : TOdbHandle;
                      const Stream      : TMemoryStream;
                      const ORMapping   : TORMapping;
                      const DatasetName : String ); virtual; abstract;

    { update an object in a dataset file with the contents of Stream. The
      Fields data member. This method need only be overriden for a driver that
      supports the damSQL data access method. Handle is related to a
      TObjectUpdate. }
    procedure InternalUpdateSQL(
                      const Handle      : TOdbHandle;
                      const Stream      : TMemoryStream;
                      const ORMapping   : TORMapping;
                      const DatasetName : String ); virtual; abstract;

    { delete the current object in a file. Positioning must be established
      to the next object if possible, if not to the previous. This method need
      only be overriden for a driver that supports the damSQL data
      access method. Handle is related to a TObjectUpdate. }
    procedure InternalDeleteSQL( const Handle      : TOdbHandle;
                                 const ORMapping   : TORMapping;
                                 const DatasetName : String ); virtual; abstract;

    { returns the current index. }
    function InternalGetIndexName( const Handle : TOdbHandle ) : String; virtual; abstract;

    { sets current index name. }
    procedure InternalSetIndexName(
                            const Handle    : TOdbHandle;
                            const IndexName : String ); virtual; abstract;

    { get the current object positioned in a file into a stream. }
    procedure InternalGetCurrent(
                          const Handle  : TOdbHandle;
                          const Stream  : TMemoryStream ); virtual; abstract;

    { lock the current object positioned. }
    function InternalLockCurrent( const Handle : TOdbHandle ) : Boolean; virtual;

    { unlock the current object positioned. }
    procedure InternalUnLockCurrent( const Handle : TOdbHandle ); virtual;

    { set the current position to the next object in a file. Returns False if
      at EOF. }
    function InternalFindNext( const Handle : TOdbHandle ) : Boolean; virtual; abstract;

    { set the current position to the previous object in a file. Returns False if
      at BOF. }
    function InternalFindPrior( const Handle : TOdbHandle ) : Boolean; virtual; abstract;

    { set the current position to the first object in a file. Returns False if
      file is empty. }
    function InternalFindFirst( const Handle : TOdbHandle ) : Boolean; virtual; abstract;

    { set the current position to the last object in a file. Returns False if
      file is empty. }
    function InternalFindLast( const Handle : TOdbHandle ) : Boolean; virtual; abstract;

    { find and position the object which is equal to the field values specified
      in Fields. Returns False if not found. }
    function InternalFindExact(
                        const Handle : TOdbHandle;
                        const Fields : TUnOrderedList;
                        const Index  : TOdbIndex ) : Boolean; virtual; abstract;

    { find and position the object which is equal to or greater than the field
      values specified in Fields. }
    procedure InternalFindNearest(
                          const Handle : TOdbHandle;
                          const Fields : TUnOrderedList;
                          const Index  : TOdbIndex ); virtual; abstract;

    { returns the number of objects in a file }
    function InternalGetObjectCount( const Handle : TOdbHandle ) : LongInt; virtual; abstract;

    function InternalGetBookmark( const Handle : TOdbHandle ) : Pointer; virtual; abstract;
    procedure InternalGotoBookmark( const Handle   : TOdbHandle;
                                    const Bookmark : Pointer ); virtual; abstract;
    procedure InternalFreeBookmark( const Handle   : TOdbHandle;
                                    const Bookmark : Pointer ); virtual; abstract;

    function InternalIsEmpty( const Handle : TOdbHandle ) : Boolean; virtual;

    procedure InternalPack( const Handle : TOdbHandle ); virtual;

    { returns the field types that are supported by the driver. }
    function GetSupportedFieldTypes : TOdbFieldTypes; virtual; abstract;

end;

TOdbDriverClass = class of TOdbDriver;

{--------------------------------------------------------------------------}

{ Object database class. A TObjectDatabase instance is required before
creating a TObjectFile instance. This class is responsible for database
operations at the database level, such as opening/closing a session and
transaction control. }
TObjectDatabase = class(TObject)
protected
    FDriver          : TOdbDriver;

    FObjectDatasets  : TUnOrderedList;

    function GetDatabaseName : String;
    function GetDatabaseOpened : Boolean;

public

    { creates an object database driver to access the database identified by
      DatabaseName. }
    constructor Create( const InOdbDriverClass : TOdbDriverClass;
                        const InDatabaseName   : String);
    destructor Destroy; override;

    property Driver : TOdbDriver read FDriver;

    { list of current TObjectFile instances belonging to the database. }
    property ObjectDatasets : TUnOrderedList read FObjectDatasets;

    property DatabaseName : String read GetDatabaseName;

    procedure StartTransaction;
    procedure Commit;
    procedure RollBack;

    procedure Open( const ServerName : String;
                    const UserName   : String;
                    const Password   : String
                  );

    procedure Close;

    property DatabaseOpened : Boolean read GetDatabaseOpened;

end;

{--------------------------------------------------------------------------}

{ Registration an Odb driver. Registered drivers may then be obtained
by using GetOdbDriver. }
procedure RegisterOdbDriver( const DriverClass  : TOdbDriverClass;
                             const DriverName   : String );

{ Returns a registered Odb Driver class. }
function GetOdbDriverClass( const DriverName : String ) : TOdbDriverClass;

{--------------------------------------------------------------------------}

{ SQL helper methods }

{ build and sql statement to created the table for ObjectDataset.
  The primary index will be added to the statement }
procedure BuildCreateTableStatement( const Statement     : TStrings;
                                     const ORMapping     : TORMapping;
                                     const DatasetName   : String;
                                     const StringType    : String;
                                     const IntegerType   : String;
                                     const FloatType     : String;
                                     const BlobType      : String );

{--------------------------------------------------------------------------}

{ build a list of create index statements (one per line), for each
  index that is not the primary index }
procedure BuildCreateIndexStatements( const Statements    : TStrings;
                                      const ObjectDataset : TObjectDataset );

{--------------------------------------------------------------------------}

{ build a query statement that will select the object identified by the
  contents in Fields and the primary key of ObjectQuery }
procedure BuildSelectObjectQuery( const SQL         : TStrings;
                                  const ORMapping   : TORMapping;
                                  const DatasetName : String );

{--------------------------------------------------------------------------}

{ build a where clause to select a singleton given the contents of fields and
  the primary index of ObjectDataset }
procedure BuildSelectWhereClause( var   WhereClause   : String;
                                  const ORMapping     : TORMapping;
                                  const DatasetName   : String );

{--------------------------------------------------------------------------}

{ build a delete statement that will delete the object identified by the
  contents in Fields and the primary key of ObjectQuery }
procedure BuildDeleteStatement( const SQL           : TStrings;
                                const ORMapping     : TORMapping;
                                const DatasetName   : String );

{--------------------------------------------------------------------------}

{ build an insert statement to insert a record identified by the
  contents of Fields. Note that this statement will not insert the
  object (ie, the blob field ) }
procedure BuildInsertStatement( const SQL           : TStrings;
                                const Fields        : TUnOrderedList;
                                const DatasetName   : String );

{--------------------------------------------------------------------}

{$ifdef __DebugWindow}
procedure SendDebugSQL( const SQL : TStrings );
{$endif}

{--------------------------------------------------------------------------}

type

EOdbError                        = class(Exception);
EOdbTransactionNotSupported      = class(Exception);
EOdbFileLockingNotSupported      = class(Exception);
EOdbObjectLockingNotSupported    = class(Exception);
EOdbFieldTypeNotSupported        = class(Exception);
EOdbInvalidFieldType             = class(Exception);
EOdbInvalidObjectFileHandle      = class(Exception);
EOdbDatabaseNotOpened            = class(Exception);
EOdbObjectFileNotPositioned      = class(Exception);
EOdbFieldsNotComparable          = class(Exception);
EOdbFieldDoesNotExist            = class(Exception);
EOdbIndexDoesNotExist            = class(Exception);
EOdbDatabaseDoesNotExist         = class(Exception);
EOdbObjectFileDoesNotExist       = class(Exception);
EOdbKeyViolation                 = class(Exception);
EOdbDriverAlreadyRegistered      = class(Exception);
EOdbDriverNotRegistered          = class(Exception);
EOdbUnsupportedAccessMethod      = class(Exception);
EOdbTableMustHavePrimaryIndex    = class(Exception);
EOdbObjectNotFound               = class(Exception);
EOdbSelectStatementNotSpecified  = class(Exception);

{--------------------------------------------------------------------}

implementation

uses
    {$ifdef __Evaluation}
      {$ifdef Win32}
      Windows,  { for HWnd }
      {$else}
      WinTypes, { for HWnd }
      WinProcs,
      {$endif}
      Dialogs,  { for MessageDlg }
    {$endif}
    ObjType,  { for TArrayIndex, EInvalidObjectType }
    ObjCmp,   { for TComparison, CompareInteger }
    ObjBuckt, { for TStringCombo }
    ObjPerst, { for TObjectStream, CloneObj }
    OdbMisc;  { for RaiseOdbException }

{--------------------------------------------------------------------}

procedure SendDebugSQL( const SQL : TStrings );
var
    LineNo : Integer;
begin
    for LineNo := 0 to (SQL.Count-1) do
        DebugOutputStr( SQL[LineNo] );
end;

{--------------------------------------------------------------------------}

var

OdbDriverRegistry : TUnOrderedList;

{--------------------------------------------------------------------------}

function CompareOdbDriverEntry(const Lhs, Rhs : TObject) : TComparison; far;
begin
    Result := CompareStringComboNoCase( Lhs, Rhs );
end;

{--------------------------------------------------------------------------}

procedure RegisterOdbDriver( const DriverClass  : TOdbDriverClass;
                             const DriverName   : String );
var
    StringCombo : TStringCombo;
begin
    StringCombo := TStringCombo.Create(DriverName, TObject(DriverClass));
    StringCombo.Ownership := False;

    if OdbDriverRegistry.Search(StringCombo, 1) then
    begin
        StringCombo.Free;
        RaiseOdbException(EOdbDriverAlreadyRegistered, DriverName);
    end
    else
        OdbDriverRegistry.Insert(StringCombo);
end;

{--------------------------------------------------------------------------}

function GetOdbDriverClass( const DriverName : String ) : TOdbDriverClass;
var
    StringCombo : TStringCombo;
begin
    Result := nil; { prevent compiler warning }
    
    StringCombo := TStringCombo.Create(DriverName, nil);
    try
        if OdbDriverRegistry.Search(StringCombo, 1) then
            Result := TOdbDriverClass(
                     (OdbDriverRegistry.CurrentObj as TStringCombo).Obj)
        else
            RaiseOdbException(EOdbDriverNotRegistered, DriverName);
    finally
        StringCombo.Free;
    end;
end;

{--------------------------------------------------------------------------}

constructor THandleData.Create( const InHandle        : LongInt;
                                       const InDataObj       : TObject );
begin
    inherited Create;

    FDataObj       := InDataObj;
    FHandle        := InHandle;
end;

{--------------------------------------------------------------------------}

destructor THandleData.Destroy;
begin
    if DataObj <> nil then
    begin
        DataObj.Free;
        DataObj := nil;
    end;

    inherited Destroy;
end;

{--------------------------------------------------------------------------}

function CompareHandleData(const Lhs, Rhs : TObject) : TComparison; far;
var
    Left  : THandleData absolute Lhs;
    Right : THandleData absolute Rhs;
begin
    Result := CompareInteger(Left.Handle, Right.Handle);
end;

{--------------------------------------------------------------------------}

constructor TOdbBookmark.Create( const ObjectFile : TObjectFile );
begin
    inherited Create;
    FObjectFile    := ObjectFile;
    FIndexName     := ObjectFile.IndexName;
    FBookmarkData  := ObjectFile.Driver.GetBookmark(ObjectFile.Handle);
end;

{--------------------------------------------------------------------------}

destructor TOdbBookmark.Destroy;
begin
    FObjectFile.Driver.FreeBookmark(FObjectFile.Handle, FBookmarkData);
    inherited Destroy;
end;

{--------------------------------------------------------------------------}

constructor TObjectDataset.Create(
                    const InDatabase        : TObjectDatabase;
                    const InDatasetName     : String;
                    const InORMappingClass  : TORMappingClass );
begin
    inherited Create;

    FCurrentObj     := nil;
    FDatabase       := InDatabase;
    FDatasetName    := InDatasetName;

    FORMapping      := InORMappingClass.Create( InDatabase );

    FDatabase.ObjectDatasets.Insert(Self);
end;

{--------------------------------------------------------------------------}

destructor TObjectDataset.Destroy;
begin
    if Opened then
        Self.Close;

    FDatabase.ObjectDatasets.Delete(Self, 1);

    FORMapping.Free;

    if FCurrentObj <> nil then
    begin
        FCurrentObj.Free;
        FCurrentObj := nil;
    end;

    inherited Destroy;
end;

{-------------------------------------------------------------------------}

function TObjectDataset.GetObjectType : TClass;
begin
    Result := FORMapping.GetObjectType;
end;

{-------------------------------------------------------------------------}

function TObjectDataset.GetFields : TUnOrderedList;
begin
    Result := FORMapping.FFields;
end;

{-------------------------------------------------------------------------}

function TObjectDataset.GetIndices : TUnOrderedList;
begin
    Result := FORMapping.FIndices;
end;

{-------------------------------------------------------------------------}

procedure TObjectDataset.CreateDataset;
begin
    Driver.CreateDataset(Self);
end;

{-------------------------------------------------------------------------}

procedure TObjectDataset.DeleteDataset;
begin
    if Opened then
        Self.Close;

    Driver.DeleteDataset(Self );
end;

{-------------------------------------------------------------------------}

function TObjectDataset.GetOpened : Boolean;
begin
    Result := (Handle <> 0);
end;

{-------------------------------------------------------------------------}

procedure TObjectDataset.Open;
begin
    if not Opened then
    begin
        BeforeOpen;

        FHandle := Driver.OpenDataset(Self);

        AfterOpen;

        if Self is TObjectFile then
        begin
            if Indices.GotoFirst then
                TObjectFile(Self).IndexName := (Indices.CurrentObj as TOdbIndex).Name;
        end;
        FindFirst;
    end;
end;

{-------------------------------------------------------------------------}

procedure TObjectDataset.Close;
begin
    if Opened then
    begin
        BeforeClose;
        Driver.CloseDataset(Handle);
        FHandle := 0;
        AfterClose;
    end;
end;

{-------------------------------------------------------------------------}

procedure TObjectDataset.CheckObjectType( const Obj : TObject );
begin
    if not (Obj is ObjectType) then
        RaiseInvalidObjectType;
end;

{-------------------------------------------------------------------------}

function TObjectDataset.GetDriver : TOdbDriver;
begin
    Result := Database.Driver;
end;

{-------------------------------------------------------------------------}

function TObjectDataset.GetCurrentObj : TObject;
begin
    if FCurrentObj = nil then
        RaiseOdbException( EOdbObjectFileNotPositioned, '' );

    Result := FCurrentObj;
end;

{-------------------------------------------------------------------------}

function TObjectDataset.FieldByName( const Name : String ) : TOdbField;
begin
    Result := GetFieldInList(Name, Fields);
end;

{-------------------------------------------------------------------------}

procedure TObjectDataset.ReloadCurrent;
var
    Stream       : TMemoryStream;
    ObjectStream : TObjectStream;
begin
    Stream       := TMemoryStream.Create;
    try
        Driver.GetCurrent(Handle, Stream);
    except
        { only free Stream explicitely if an error occurs. }
        Stream.Free;
        raise;
    end;

    Stream.Seek(0, soFromBeginning);
    FORMapping.AfterObjectRead(Stream);

    Stream.Seek(0, soFromBeginning);
    ObjectStream := TObjectStream.Create(Stream);
    try
        if FCurrentObj <> nil then
        begin
            FCurrentObj.Free;
            FCurrentObj := nil;
        end;

        FCurrentObj := ObjectStream.ReadObject;

        FORMapping.UpdateFieldsFromObj(CurrentObj);
    finally
        { Stream is destroyed by the ObjectStream itself.}
        ObjectStream.Free;
    end;
end;

{-------------------------------------------------------------------------}

procedure TObjectDataset.GetAllUniqueIndices( const IndexList : TUnOrderedList );
var
    Index : TOdbIndex;
begin
    IndexList.Ownership := False;
    IndexList.Clear;

    if Indices.GotoFirst then
    repeat
        Index := Indices.CurrentObj as TOdbIndex;

        if (iaUnique in Index.Attributes) or
           (iaPrimary in Index.Attributes) then
            IndexList.Insert(Index);
    until not Indices.GotoNext;
end;

{-------------------------------------------------------------------------}

function TORMapping.GetPrimaryIndex( const DatasetName : String ) : TOdbIndex;
var
    Index : TOdbIndex;

begin
    Result := nil;

    if Indices.GotoFirst then
    repeat
        Index := Indices.CurrentObj as TOdbIndex;

        if iaPrimary in Index.Attributes then
        begin
            Result := Index;
            Exit;
        end;
    until not Indices.GotoNext;

    { if reach here, primary index not found }
    RaiseOdbException( EOdbTableMustHavePrimaryIndex, DatasetName );
end;

{-------------------------------------------------------------------------}

procedure TObjectDataset.GetModifiedUniqueIndices(
                                        const IndexList      : TUnOrderedList;
                                        const PreviousFields : TUnOrderedList;
                                        const CurrentFields  : TUnOrderedList );

    { returns True if the field values in PreviousFields and CurrentFields
      are different for the fields contained in FieldNames. }
    function FieldsAreModified( const FieldNames     : TUnOrderedList;
                                const PreviousFields : TUnOrderedList;
                                const CurrentFields  : TUnOrderedList
                              ) : Boolean;
    var
        StringCombo   : TStringCombo;
        PreviousField : TOdbField;
        CurrentField  : TOdbField;
    begin
        Result := False;
        if FieldNames.GotoFirst then
        repeat
            StringCombo   := FieldNames.CurrentObj as TStringCombo;
            PreviousField := GetFieldInList(StringCombo.Str, PreviousFields);
            CurrentField  := GetFieldInList(StringCombo.Str, CurrentFields);

            if not FieldsEqual(PreviousField, CurrentField) then
            begin
                Result := True;
                Exit;
            end;

        until not FieldNames.GotoNext;
    end;
var
    Index : TOdbIndex;

begin
    IndexList.Ownership := False;
    IndexList.Clear;

    if Indices.GotoFirst then
    repeat
        Index := Indices.CurrentObj as TOdbIndex;

        if (
            (iaUnique in Index.Attributes) or
            (iaPrimary in Index.Attributes)
           )
           and
           FieldsAreModified( Index.FieldNames,
                              PreviousFields, CurrentFields ) then
            IndexList.Insert(Index);

    until not Indices.GotoNext;

end;

{-------------------------------------------------------------------------}

function TObjectDataset.IsEmpty : Boolean;
begin
    Result := Driver.IsEmpty(Handle);
end;

{-------------------------------------------------------------------------}

function TObjectDataset.FindNext : Boolean;
begin
    Result := Driver.FindNext(Handle);

    if Result then
        ReloadCurrent;
end;

{-------------------------------------------------------------------------}

function TObjectDataset.FindPrior : Boolean;
begin
    Result := Driver.FindPrior(Handle);

    if Result then
        ReloadCurrent;
end;

{-------------------------------------------------------------------------}

function TObjectDataset.FindFirst : Boolean;
begin
    Result := Driver.FindFirst(Handle);

    if Result then
        ReloadCurrent;
end;

{-------------------------------------------------------------------------}

function TObjectDataset.FindLast : Boolean;
begin
    Result := Driver.FindLast(Handle);

    if Result then
        ReloadCurrent;
end;

{-------------------------------------------------------------------------}

procedure TObjectDataset.BeforeInsert;
begin
    FORMapping.BeforeInsert;
end;

{-------------------------------------------------------------------------}

procedure TObjectDataset.AfterInsert;
begin
    FORMapping.AfterInsert;
end;

{-------------------------------------------------------------------------}

procedure TObjectDataset.BeforeUpdate;
begin
    FORMapping.BeforeUpdate;
end;

{-------------------------------------------------------------------------}

procedure TObjectDataset.AfterUpdate;
begin
    FORMapping.AfterUpdate;
end;

{-------------------------------------------------------------------------}

procedure TObjectDataset.BeforeDelete;
begin
    FORMapping.BeforeDelete;
end;

{-------------------------------------------------------------------------}

procedure TObjectDataset.AfterDelete;
begin
    FORMapping.AfterDelete;
end;

{-------------------------------------------------------------------------}

procedure TObjectDataset.BeforeObjectPost( const Stream  : TMemoryStream );
begin
    FORMapping.BeforeObjectPost(Stream);
end;

{-------------------------------------------------------------------------}

procedure TObjectDataset.AfterObjectRead( const Stream : TMemoryStream );
begin
    FORMapping.AfterObjectRead(Stream);
end;

{-------------------------------------------------------------------------}

procedure TObjectDataset.BeforeOpen;
begin
end;

{-------------------------------------------------------------------------}

procedure TObjectDataset.AfterOpen;
begin
end;

{-------------------------------------------------------------------------}

procedure TObjectDataset.BeforeClose;
begin
end;

{-------------------------------------------------------------------------}

procedure TObjectDataset.AfterClose;
begin
end;

{-------------------------------------------------------------------------}

constructor TObjectFile.Create( const InDatabase        : TObjectDatabase;
                                const InDatasetName     : String;
                                const InORMappingClass  : TORMappingClass );
begin
    InDatabase.Driver.CheckAccessMethod(damISAM);

    inherited Create(InDatabase, InDatasetName, InORMappingClass);
end;

{-------------------------------------------------------------------------}

type

{ OR mapping class to support the Spider ODB version 1 interface }
TVersion1ORMapping = class(TORMapping)
protected
    FObjectFile : TObjectFile;
    FObjectType : TClass;

    function GetObjectType : TClass; override;

    procedure DefineFields; override;
    procedure DefineIndices; override;

public

    constructor Create( const InDatabase   : TObjectDatabase ); override;

    property ObjectFile : TObjectFile read FObjectFile
                                      write FObjectFile;

    property ObjectType : TClass read FObjectType
                                 write FObjectType;

    procedure UpdateFieldsFromObj( const Obj       : TObject ); override;

    procedure UpdateObjFromFields( const Obj       : TObject ); override;

    procedure BeforeInsert; override;
    procedure AfterInsert; override;

    procedure BeforeUpdate; override;
    procedure AfterUpdate; override;

    procedure BeforeDelete; override;
    procedure AfterDelete; override;

    procedure BeforeObjectPost( const Stream  : TMemoryStream ); override;
    procedure AfterObjectRead( const Stream : TMemoryStream ); override;
end;

{-------------------------------------------------------------------------}

constructor TVersion1ORMapping.Create( const InDatabase   : TObjectDatabase );
begin
    inherited Create(InDatabase);
end;

{-------------------------------------------------------------------------}

function TVersion1ORMapping.GetObjectType : TClass;
begin
    Result := FObjectType;
end;

{-------------------------------------------------------------------------}

procedure TVersion1ORMapping.DefineFields;
begin
    if Assigned(FObjectFile) then
        FObjectFile.DefineFields;
end;

{-------------------------------------------------------------------------}

procedure TVersion1ORMapping.DefineIndices;
begin
    if Assigned(FObjectFile) then
        FObjectFile.DefineIndices;
end;

{-------------------------------------------------------------------------}

procedure TVersion1ORMapping.UpdateFieldsFromObj( const Obj       : TObject );
begin
    FObjectFile.UpdateFieldsFromObj(Obj);
end;

{-------------------------------------------------------------------------}

procedure TVersion1ORMapping.UpdateObjFromFields( const Obj       : TObject );
begin
    FObjectFile.UpdateObjFromFields(Obj);
end;

{-------------------------------------------------------------------------}

procedure TVersion1ORMapping.BeforeInsert;
begin
    FObjectFile.BeforeInsert;
end;

{-------------------------------------------------------------------------}

procedure TVersion1ORMapping.AfterInsert;
begin
    FObjectFile.AfterInsert;
end;

{-------------------------------------------------------------------------}

procedure TVersion1ORMapping.BeforeUpdate;
begin
    FObjectFile.BeforeUpdate;
end;

{-------------------------------------------------------------------------}

procedure TVersion1ORMapping.AfterUpdate;
begin
    FObjectFile.AfterUpdate;
end;

{-------------------------------------------------------------------------}

procedure TVersion1ORMapping.BeforeDelete;
begin
    FObjectFile.BeforeDelete;
end;

{-------------------------------------------------------------------------}

procedure TVersion1ORMapping.AfterDelete;
begin
    FObjectFile.AfterDelete;
end;

{-------------------------------------------------------------------------}

procedure TVersion1ORMapping.BeforeObjectPost( const Stream  : TMemoryStream );
begin
    FObjectFile.BeforeObjectPost(Stream);
end;

{-------------------------------------------------------------------------}

procedure TVersion1ORMapping.AfterObjectRead( const Stream : TMemoryStream );
begin
    FObjectFile.AfterObjectRead(Stream);
end;

{-------------------------------------------------------------------------}

constructor TObjectFile.CreateVersion1( const InDatabase       : TObjectDatabase;
                                        const InObjectFileName : String;
                                        const InObjectType     : TClass );
begin
    inherited Create( InDatabase, InObjectFileName, TVersion1ORMapping );

    (FORMapping as TVersion1ORMapping).ObjectType := InObjectType;
    (FORMapping as TVersion1ORMapping).ObjectFile := Self;

    { call DefineFields and DefineIndices explicitely, because when the
      TORMapping instance was created, the ObjectFile property was not
      assigned. }
    FORMapping.DefineFields;
    FORMapping.DefineIndices;
end;

{-------------------------------------------------------------------------}

procedure TObjectFile.SetCurrentObj( const Obj : TObject );
begin
    if Obj <> nil then
         CheckObjectType(Obj);

    if FCurrentObj <> nil then
    begin
        FCurrentObj.Free;
        FCurrentObj := nil;
    end;

    if Obj <> nil then
        FCurrentObj := CloneObj(Obj);
end;

{-------------------------------------------------------------------------}

function TObjectFile.AddField(
                    const InName       : String;
                    const InType       : TOdbFieldType;
                    const InSize       : Byte ) : TOdbField;
begin
    Result := FORMapping.AddField(InName, InType, InSize);
end;

{-------------------------------------------------------------------------}

function TObjectFile.AddIndex(
                    const IndexName  : String;
                    const Attributes : TOdbIndexAttributes;
                    const FieldNames : array of String ) : TOdbIndex;
begin
    Result := FORMapping.AddIndex(IndexName, Attributes, FieldNames);
end;

{-------------------------------------------------------------------------}

procedure TObjectFile.UpdateFieldsFromObj( const Obj    : TObject );
begin
end;

{-------------------------------------------------------------------------}

procedure TObjectFile.DefineFields;
begin
end;

{-------------------------------------------------------------------------}

procedure TObjectFile.DefineIndices;
begin
end;

{-------------------------------------------------------------------------}

procedure TObjectFile.CreateFile;
begin
    CreateDataset;
end;

{-------------------------------------------------------------------------}

procedure TObjectFile.DeleteFile;
begin
    DeleteDataset;
end;

{-------------------------------------------------------------------------}

procedure TObjectFile.Insert;
var
    Stream       : TMemoryStream;
    ObjectStream : TObjectStream;
    IndexList    : TUnOrderedList;
    //??PrevFields   : TUnOrderedList;
begin
    FORMapping.BeforeInsert;

    if Driver.CheckDuplicateInsert and
       not Driver.IsEmpty(Handle) then
    begin
        (*
        PrevFields := CloneObj(Fields) as TUnOrderedList;
        try
            FORMapping.UpdateFieldsFromObj(CurrentObj);
            IndexList := TUnOrderedList.Create( TOdbIndex, CompareOdbIndex );
            try
                GetAllUniqueIndices(IndexList);
                if (IndexList.Size > 0) then
                begin
                    try
                        CheckForDuplicate(IndexList, Fields);
                    except
                        FFields.Free;
                        FFields := CloneObj(PrevFields) as TUnOrderedList;
                        raise;
                    end;
                end;
            finally
                IndexList.Free;
            end;
        finally
            PrevFields.Free;
        end;
        *)
        FORMapping.UpdateFieldsFromObj(CurrentObj);
        IndexList := TUnOrderedList.Create( TOdbIndex, CompareOdbIndex );
        try
            GetAllUniqueIndices(IndexList);
            if (IndexList.Size > 0) then
            begin
                CheckForDuplicate(IndexList, Fields);
            end;
        finally
            IndexList.Free;
        end;

    end
    else
        FORMapping.UpdateFieldsFromObj(CurrentObj);

    Stream       := TMemoryStream.Create;
    ObjectStream := TObjectStream.Create(Stream);
    try
        if FieldListHasAutoInc(Fields) then
        begin
            { if the object file has an autoincrement field, first insert the
              fields, then update the object to ensure the object is stored
              with the proper value in the autoincrement field. }
            Driver.Insert(Handle, Fields, Stream);

            FORMapping.UpdateObjFromFields(CurrentObj);

            ObjectStream.WriteObject(CurrentObj);
            Stream.Seek(0, soFromBeginning);

            FORMapping.BeforeObjectPost(Stream);
            Stream.Seek(0, soFromBeginning);
            Driver.Update(Handle, Fields, Stream);
        end
        else
        begin
            ObjectStream.WriteObject(CurrentObj);
            Stream.Seek(0, soFromBeginning);

            FORMapping.BeforeObjectPost(Stream);
            Stream.Seek(0, soFromBeginning);
            Driver.Insert(Handle, Fields, Stream);
        end;

    finally
        { Stream is destroyed by the ObjectStream itself.}
        ObjectStream.Free;
    end;

    FORMapping.AfterInsert;
end;

{-------------------------------------------------------------------------}

procedure TObjectFile.Update;
var
    Stream       : TMemoryStream;
    ObjectStream : TObjectStream;
    IndexList    : TUnOrderedList;
    PrevFields   : TUnOrderedList;

begin
    if FCurrentObj = nil then
        RaiseOdbException( EOdbObjectFileNotPositioned, '');

    FORMapping.BeforeUpdate;

    if Driver.CheckDuplicateUpdate  and
       (Self is TObjectFile) then
    begin
        (*
        PrevFields := CloneObj(Fields) as TUnOrderedList;
        try
            FORMapping.UpdateFieldsFromObj(CurrentObj);
            IndexList := TUnOrderedList.Create( TOdbIndex, CompareOdbIndex );
            try
                GetModifiedUniqueIndices(IndexList, PrevFields, Fields);
                if (IndexList.Size > 0) then
                begin
                    try
                        CheckForDuplicate(IndexList, Fields);
                    except
                        FFields.Free;
                        FFields := CloneObj(PrevFields) as TUnOrderedList;
                        raise;
                    end;
                end;
            finally
                IndexList.Free;
            end;
        finally
            PrevFields.Free;
        end;
        *)
        PrevFields := CloneObj(Fields) as TUnOrderedList;
        try
            FORMapping.UpdateFieldsFromObj(CurrentObj);
            IndexList := TUnOrderedList.Create( TOdbIndex, CompareOdbIndex );
            try
                GetModifiedUniqueIndices(IndexList, PrevFields, Fields);
                if (IndexList.Size > 0) then
                begin
                    CheckForDuplicate(IndexList, Fields);
                end;
            finally
                IndexList.Free;
            end;
        finally
            PrevFields.Free;
        end;
    end
    else
        FORMapping.UpdateFieldsFromObj(CurrentObj);

    Stream       := TMemoryStream.Create;
    ObjectStream := TObjectStream.Create(Stream);
    try
        ObjectStream.WriteObject(CurrentObj);
        Stream.Seek(0, soFromBeginning);

        FORMapping.BeforeObjectPost(Stream);
        Stream.Seek(0, soFromBeginning);

        Driver.Update(Handle, Fields, Stream);
    finally
        { Stream is destroyed by the ObjectStream itself.}
        ObjectStream.Free;
    end;

    FORMapping.AfterUpdate;
end;

{-------------------------------------------------------------------------}

procedure TObjectFile.Delete;
begin
    if FCurrentObj = nil then
        RaiseOdbException( EOdbObjectFileNotPositioned, '' );

    FORMapping.BeforeDelete;
    Driver.Delete(Handle);
    FORMapping.AfterDelete;

    if Driver.IsEmpty(Handle) then
        CurrentObj := nil
    else
        ReloadCurrent;
end;

{-------------------------------------------------------------------------}

function TObjectFile.GetCurrentIndex : TOdbIndex;
var
    Key     : TOdbIndex;
begin
    Key := TOdbIndex.Create(IndexName, []);
    try
        if not Indices.Search(Key, 1) then
            RaiseOdbException( EOdbIndexDoesNotExist, IndexName);
    finally
       Key.Free;
    end;

    Result := Indices.CurrentObj as TOdbIndex;
end;

{-------------------------------------------------------------------------}

function TObjectFile.GetObjectCount : LongInt;
begin
    Result := Driver.GetObjectCount(Handle);
end;

{-------------------------------------------------------------------------}

function TObjectFile.LockCurrent : Boolean;
begin
    Result := Driver.LockCurrent(Handle);
end;

{-------------------------------------------------------------------------}

procedure TObjectFile.UnLockCurrent;
begin
    Driver.UnLockCurrent(Handle);
end;

{-------------------------------------------------------------------------}

procedure TObjectFile.UpdateObjFromFields( const Obj    : TObject );
begin
end;

{-------------------------------------------------------------------------}

function TObjectFile.GetIndexName : String;
begin
    Result := Driver.GetIndexName(Handle)
end;

{-------------------------------------------------------------------------}

procedure TObjectFile.SetIndexName( const InIndexName : String );
begin
    if IndexName <> InIndexName then
        Driver.SetIndexName(Handle, InIndexName);
end;

{-------------------------------------------------------------------------}

function TObjectFile.FindExact : Boolean;
begin
    Result := False;

    if FCurrentObj = nil then
        Exit;

    FORMapping.UpdateFieldsFromObj(CurrentObj);

    Result := Driver.FindExact(Handle, Fields, CurrentIndex);

    if Result then
        ReloadCurrent;
end;

{-------------------------------------------------------------------------}

procedure TObjectFile.FindNearest;
begin
    if FCurrentObj = nil then
        Exit;

    FORMapping.UpdateFieldsFromObj(CurrentObj);

    Driver.FindNearest(Handle, Fields, CurrentIndex);

    ReloadCurrent;
end;

{-------------------------------------------------------------------------}

function TObjectFile.GetBookmark : TOdbBookmark;
begin
    Result := TOdbBookmark.Create(Self);
end;

{-------------------------------------------------------------------------}

procedure TObjectFile.GotoBookmark( const Bookmark     : TOdbBookmark;
                                    const FreeBookmark : Boolean );
begin
    if IndexName <> Bookmark.IndexName then
        IndexName := Bookmark.IndexName;

    Driver.GotoBookmark(Handle, Bookmark.FBookmarkData);

    ReloadCurrent;

    if FreeBookmark then
        Bookmark.Free;
end;

{-------------------------------------------------------------------------}

procedure TObjectFile.Pack;
begin
    Driver.Pack(Handle);
end;

{-------------------------------------------------------------------------}

procedure TObjectFile.OpenExclusive;
begin
    if not Opened then
    begin
        BeforeOpen;
        FHandle := Driver.OpenExclusive(Self);
        AfterOpen;

        if Indices.GotoFirst then
            IndexName := (Indices.CurrentObj as TOdbIndex).Name;

        FindFirst;
    end;
end;

{-------------------------------------------------------------------------}

procedure TObjectFile.CheckForDuplicate(
                             const InIndexList : TUnOrderedList;
                             const InFields    : TUnOrderedList );
var
    Bookmark   : TOdbBookmark;
    Index      : TOdbIndex;
begin
    Bookmark := GetBookmark;
    try
        if InIndexList.GotoFirst then
        repeat
            Index     := InIndexList.CurrentObj as TOdbIndex;
            IndexName := Index.Name;

            if Driver.FindExact(Handle, InFields, Index) then
                RaiseOdbException( EOdbKeyViolation, Index.Name );
        until not InIndexList.GotoNext;
    finally
        IndexName := Bookmark.IndexName;
        { do not use Self.GotoBookmark to prevent reloading the current
          object. }
        Driver.GotoBookmark(Handle, Bookmark.FBookmarkData);
        Bookmark.Free;
    end;
end;                

{-------------------------------------------------------------------------}

constructor TObjectQuery.Create( const InDatabase       : TObjectDatabase;
                                 const InDatasetName    : String;
                                 const InORMappingClass : TORMappingClass );
begin
    InDatabase.Driver.CheckAccessMethod(damSQL);

    inherited Create( InDatabase,
                      InDatasetName,
                      InORMappingClass );

    FSelect := TStringList.Create;
end;

{-------------------------------------------------------------------------}

destructor TObjectQuery.Destroy;
begin
    FSelect.Free;

    inherited Destroy;
end;

{-------------------------------------------------------------------------}

function TObjectQuery.Search( const Key      : TObject ) : TObject;
begin
    if Self.Opened then
        Self.Close;

    Self.ORMapping.UpdateFieldsFromObj(Key);

    BuildSelectObjectQuery( Self.Select,
                            Self.ORMapping,
                            Self.DatasetName );

    Result := nil;
    Self.Open;
    try
        if not Self.IsEmpty then
            Result := Self.CurrentObj;
    finally
        Self.Close;
    end;
end;

{-------------------------------------------------------------------------}

constructor TObjectUpdate.Create( const InDatabase       : TObjectDatabase;
                                  const InDatasetName    : String;
                                  const InORMappingClass : TORMappingClass );
begin
    inherited Create;

    FDatabase       := InDatabase;
    FDatasetName    := InDatasetName;
    FORMapping      := InORMappingClass.Create(InDatabase);

    FHandle         := FDatabase.Driver.InternalCreateObjectUpdate;
end;

{-------------------------------------------------------------------------}

destructor TObjectUpdate.Destroy;
begin
    FDatabase.Driver.InternalDestroyObjectUpdate(FHandle);

    FORMapping.Free;

    inherited Destroy;
end;

{-------------------------------------------------------------------------}

procedure TObjectUpdate.Insert( const Obj : TObject );
var
    Stream       : TMemoryStream;
    ObjectStream : TObjectStream;
begin
    FORMapping.BeforeInsert;

    FORMapping.UpdateFieldsFromObj(Obj);

    Stream       := TMemoryStream.Create;
    ObjectStream := TObjectStream.Create(Stream);
    try
        ObjectStream.WriteObject(Obj);
        Stream.Seek(0, soFromBeginning);

        FORMapping.BeforeObjectPost(Stream);
        Stream.Seek(0, soFromBeginning);

        FDatabase.Driver.InsertSQL(FHandle, Stream, FORMapping, FDatasetName);

    finally
        { Stream is destroyed by the ObjectStream itself.}
        ObjectStream.Free;
    end;

    FORMapping.AfterInsert;
end;

{-------------------------------------------------------------------------}

procedure TObjectUpdate.Delete( const Obj : TObject );
begin
    FORMapping.BeforeDelete;

    FORMapping.UpdateFieldsFromObj(Obj);

    FDatabase.Driver.DeleteSQL(FHandle, FORMapping, FDatasetName);

    FORMapping.AfterDelete;
end;

{-------------------------------------------------------------------------}

procedure TObjectUpdate.Update( const Obj : TObject );
var
    Stream       : TMemoryStream;
    ObjectStream : TObjectStream;

begin
    FORMapping.BeforeUpdate;

    FORMapping.UpdateFieldsFromObj(Obj);

    Stream       := TMemoryStream.Create;
    ObjectStream := TObjectStream.Create(Stream);
    try
        ObjectStream.WriteObject(Obj);
        Stream.Seek(0, soFromBeginning);

        FORMapping.BeforeObjectPost(Stream);
        Stream.Seek(0, soFromBeginning);

        FDatabase.Driver.UpdateSQL(FHandle, Stream, FORMapping, FDatasetName);

    finally
        { Stream is destroyed by the ObjectStream itself.}
        ObjectStream.Free;
    end;


    FORMapping.AfterUpdate;
end;

{-------------------------------------------------------------------------}

constructor TORMapping.Create( const InDatabase : TObjectDatabase );
begin
    inherited Create;

    FDatabase       := InDatabase;

    FFields         := TUnOrderedList.Create(TOdbField, CompareOdbField);
    FIndices        := TUnOrderedList.Create(TOdbIndex, CompareOdbIndex);

    DefineFields;
    DefineIndices;
end;

{-------------------------------------------------------------------------}

destructor TORMapping.Destroy;
begin
    FIndices.Free;
    FFields.Free;

    inherited Destroy;
end;

{-------------------------------------------------------------------------}

function TORMapping.AddField( const InName       : String;
                              const InType       : TOdbFieldType;
                              const InSize       : Byte ) : TOdbField;
begin
    if not (InType in FDatabase.Driver.SupportedFieldTypes) then
        RaiseOdbException( EOdbFieldTypeNotSupported, '' );

    Result := TOdbField.Create(InName, InType, InSize);
    FFields.InsertAtTail(Result);
end;

{-------------------------------------------------------------------------}

function TORMapping.AddIndex( const IndexName  : String;
                              const Attributes : TOdbIndexAttributes;
                              const FieldNames : array of String ) : TOdbIndex;
var
    FieldNo : Integer;
begin
    Result := TOdbIndex.Create(IndexName, Attributes);

    for FieldNo := Low(FieldNames) to High(FieldNames) do
    begin
        Result.FieldNames.InsertAtTail( TStringCombo.Create(
                                               FieldNames[FieldNo], nil));
    end;

    FIndices.InsertAtTail(Result);
end;

{-------------------------------------------------------------------------}

function TORMapping.FieldByName( const Name : String ) : TOdbField;
begin
    Result := GetFieldInList(Name, FFields);
end;

{-------------------------------------------------------------------------}

procedure TORMapping.UpdateObjFromFields( const Obj       : TObject );
begin
end;

{-------------------------------------------------------------------------}

procedure TORMapping.BeforeInsert;
begin
end;

{-------------------------------------------------------------------------}

procedure TORMapping.AfterInsert;
begin
end;

{-------------------------------------------------------------------------}

procedure TORMapping.BeforeUpdate;
begin
end;

{-------------------------------------------------------------------------}

procedure TORMapping.AfterUpdate;
begin
end;

{-------------------------------------------------------------------------}

procedure TORMapping.BeforeDelete;
begin
end;

{-------------------------------------------------------------------------}

procedure TORMapping.AfterDelete;
begin
end;

{-------------------------------------------------------------------------}

procedure TORMapping.BeforeObjectPost( const Stream  : TMemoryStream );
begin
end;

{-------------------------------------------------------------------------}

procedure TORMapping.AfterObjectRead( const Stream : TMemoryStream );
begin
end;

{-------------------------------------------------------------------------}

constructor TOdbDriver.Create(const InDatabaseName : String);
begin
    inherited Create;

    FDatabaseName := InDatabaseName;

    FHandles      := TArray.Create( THandleData, CompareHandleData,
                                    200, 50 );

    FSupportedFieldTypes := GetSupportedFieldTypes;
end;

{--------------------------------------------------------------------------}

destructor TOdbDriver.Destroy;
begin
    if DatabaseOpened then
        CloseDatabase;

    FHandles.Free;

    inherited Destroy;
end;

{--------------------------------------------------------------------------}

function TOdbDriver.GetHandleDataObj( const Handle : TOdbHandle ) : TObject;
var
    Key   : THandleData;
    Index : TArrayIndex;
begin
    Key := THandleData.Create(Handle, nil);
    try
        Index := FHandles.BinarySearch(Key);
        if Index = CInvalidArrayIndex then
            RaiseOdbException( EOdbInvalidObjectFileHandle, '' );

        Result := (FHandles[Index] as THandleData).DataObj;
    finally
        Key.Free;
    end;
end;

{--------------------------------------------------------------------------}

function TOdbDriver.NewHandle(
                                  const InDataObj       : TObject ) : TOdbHandle;
var
    HandleData : THandleData;
begin
    HandleData := THandleData.Create(1, InDataObj);

    { loop until a free handle number is found. }
    while FHandles.BinarySearch(HandleData) <> CInvalidArrayIndex do
        HandleData.Handle := HandleData.Handle + 1;

    FHandles.Insert(HandleData);
    FHandles.ShellSort;

    Result := HandleData.Handle;
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.ReleaseHandle( const Handle : TOdbHandle );
var
    Key   : THandleData;
    Index : TArrayIndex;
begin
    Key := THandleData.Create(Handle, nil);
    try
        Index := FHandles.BinarySearch(Key);
        if Index = CInvalidArrayIndex then
            RaiseOdbException( EOdbInvalidObjectFileHandle, '');

        FHandles.Delete(Index);
        FHandles.ShellSort;
    finally
        Key.Free;
    end;
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.StartTransaction;
begin
    CheckDatabaseOpened;

    InternalStartTransaction;
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.Commit;
begin
    CheckDatabaseOpened;

    InternalCommit;
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.Rollback;
begin
    CheckDatabaseOpened;

    InternalRollback;
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.OpenDatabase(
                        const ServerName : String;
                        const UserName   : String;
                        const Password   : String
                      );
begin
    InternalOpenDatabase(ServerName, UserName, Password);

    FDatabaseOpened := True;
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.CloseDatabase;
begin
    InternalCloseDatabase;

    FDatabaseOpened := False;
end;

{--------------------------------------------------------------------------}

function TOdbDriver.OpenDataset( const ObjectDataset : TObjectDataset ) : TOdbHandle;
begin
    CheckDatabaseOpened;

    Result := InternalOpenDataset(ObjectDataset);
end;

{--------------------------------------------------------------------------}

function TOdbDriver.OpenExclusive( const ObjectFile : TObjectFile ) : TOdbHandle;
begin
    CheckDatabaseOpened;

    Result := InternalOpenExclusive(ObjectFile);
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.CloseDataset( const Handle : TOdbHandle );
begin
    CheckDatabaseOpened;

    InternalCloseDataset(Handle);
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.CreateDataset( const ObjectDataset : TObjectDataset
                    );
begin
    CheckDatabaseOpened;

    InternalCreateDataset(ObjectDataset);
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.DeleteDataset( const ObjectDataset : TObjectDataset );
begin
    CheckDatabaseOpened;

    InternalDeleteDataset(ObjectDataset);
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.Insert(
                  const Handle      : TOdbHandle;
                  const Fields      : TUnOrderedList;
                  const Stream      : TMemoryStream );
begin
    CheckDatabaseOpened;

    InternalInsert(Handle, Fields, Stream);
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.Update(
                  const Handle     : TOdbHandle;
                  const Fields     : TUnOrderedList;
                  const Stream     : TMemoryStream );
begin
    CheckDatabaseOpened;

    InternalUpdate(Handle, Fields, Stream);
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.Delete( const Handle : TOdbHandle );
begin
    CheckDatabaseOpened;

    InternalDelete(Handle);
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.InsertSQL(
                  const Handle       : TOdbHandle;
                  const Stream       : TMemoryStream;
                  const ORMapping    : TORMapping;
                  const DatasetName  : String );
begin
    CheckDatabaseOpened;

    InternalInsertSQL(Handle, Stream, ORMapping, DatasetName);
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.UpdateSQL(
                  const Handle      : TOdbHandle;
                  const Stream      : TMemoryStream;
                  const ORMapping   : TORMapping;
                  const DatasetName : String );
begin
    CheckDatabaseOpened;

    InternalUpdateSQL(Handle, Stream, ORMapping, DatasetName);
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.DeleteSQL( const Handle      : TOdbHandle;
                                const ORMapping   : TORMapping;
                                const DatasetName : String );
begin
    CheckDatabaseOpened;

    InternalDeleteSQL(Handle, ORMapping, DatasetName);
end;

{--------------------------------------------------------------------------}

function TOdbDriver.GetIndexName( const Handle : TOdbHandle ) : String;
begin
    CheckDatabaseOpened;

    Result := InternalGetIndexName(Handle);
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.SetIndexName(
                        const Handle    : TOdbHandle;
                        const IndexName : String );
begin
    CheckDatabaseOpened;

    InternalSetIndexName(Handle, IndexName);
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.GetCurrent(
                      const Handle  : TOdbHandle;
                      const Stream  : TMemoryStream );
begin
    CheckDatabaseOpened;

    InternalGetCurrent(Handle, Stream);
end;

{--------------------------------------------------------------------------}

function TOdbDriver.LockCurrent( const Handle : TOdbHandle ) : Boolean;
begin
    CheckDatabaseOpened;

    Result := InternalLockCurrent(Handle);
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.UnLockCurrent( const Handle : TOdbHandle );
begin
    CheckDatabaseOpened;

    InternalUnLockCurrent(Handle);
end;

{--------------------------------------------------------------------------}

function TOdbDriver.FindNext( const Handle : TOdbHandle ) : Boolean;
begin
    CheckDatabaseOpened;

    Result := InternalFindNext(Handle);
end;

{--------------------------------------------------------------------------}

function TOdbDriver.FindPrior( const Handle : TOdbHandle ) : Boolean;
begin
    CheckDatabaseOpened;

    Result := InternalFindPrior(Handle);
end;

{--------------------------------------------------------------------------}

function TOdbDriver.FindFirst( const Handle : TOdbHandle ) : Boolean;
begin
    CheckDatabaseOpened;

    Result := InternalFindFirst(Handle);
end;

{--------------------------------------------------------------------------}

function TOdbDriver.FindLast( const Handle : TOdbHandle ) : Boolean;
begin
    CheckDatabaseOpened;

    Result := InternalFindLast(Handle);
end;

{--------------------------------------------------------------------------}

function TOdbDriver.FindExact(
                    const Handle : TOdbHandle;
                    const Fields : TUnOrderedList;
                    const Index  : TOdbIndex ) : Boolean;
begin
    CheckDatabaseOpened;

    Result := InternalFindExact(Handle, Fields, Index);
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.FindNearest(
                      const Handle : TOdbHandle;
                      const Fields : TUnOrderedList;
                      const Index  : TOdbIndex );
begin
    CheckDatabaseOpened;

    InternalFindNearest(Handle, Fields, Index);
end;

{--------------------------------------------------------------------------}

function TOdbDriver.GetBookmark( const Handle : TOdbHandle ) : Pointer;
begin
    CheckDatabaseOpened;

    Result := InternalGetBookmark(Handle);
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.GotoBookmark( const Handle   : TOdbHandle;
                                   const Bookmark : Pointer );
begin
    CheckDatabaseOpened;

    InternalGotoBookmark(Handle, Bookmark);
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.FreeBookmark( const Handle   : TOdbHandle;
                                   const Bookmark : Pointer );
begin
    CheckDatabaseOpened;

    InternalFreeBookmark(Handle, Bookmark);
end;

{--------------------------------------------------------------------------}

function TOdbDriver.GetObjectCount( const Handle : TOdbHandle ) : LongInt;
begin
    CheckDatabaseOpened;

    Result := InternalGetObjectCount(Handle);
end;

{--------------------------------------------------------------------------}

function TOdbDriver.IsEmpty( const Handle : TOdbHandle ) : Boolean;
begin
    CheckDatabaseOpened;

    Result := InternalIsEmpty(Handle);
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.Pack( const Handle : TOdbHandle );
begin
    CheckDatabaseOpened;

    InternalPack(Handle);
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.InternalOpenDatabase(
                        const ServerName : String;
                        const UserName   : String;
                        const Password   : String
                      );
begin
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.InternalCloseDatabase;
begin
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.CheckDatabaseOpened;
begin
    if not DatabaseOpened then
        RaiseOdbException( EOdbDatabaseNotOpened, '' );
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.CheckAccessMethod( const AccessMethod : TDataAccessMethod );
begin
    if not (AccessMethod in GetDataAccessMethods) then
        RaiseOdbException(
                           EOdbUnsupportedAccessMethod,
                           IntToStr(Ord(AccessMethod)));
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.InternalStartTransaction;
begin
    RaiseOdbException( EOdbTransactionNotSupported, '' );
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.InternalCommit;
begin
    RaiseOdbException( EOdbTransactionNotSupported, '' );
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.InternalRollback;
begin
    RaiseOdbException( EOdbTransactionNotSupported, '' );
end;

{--------------------------------------------------------------------------}

function TOdbDriver.InternalOpenExclusive( const ObjectFile : TObjectFile ) : TOdbHandle;
begin
    Result := 0; { prevent compiler warning }

    RaiseOdbException( EOdbFileLockingNotSupported, '' );
end;

{--------------------------------------------------------------------------}

function TOdbDriver.InternalLockCurrent( const Handle : TOdbHandle ) : Boolean;
begin
    { prevent compiler warning. }
    Result := False;
    
    RaiseOdbException( EOdbObjectLockingNotSupported, '' );
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.InternalUnLockCurrent( const Handle : TOdbHandle );
begin
    RaiseOdbException( EOdbObjectLockingNotSupported, '' );
end;

{--------------------------------------------------------------------------}

function TOdbDriver.InternalIsEmpty( const Handle : TOdbHandle ) : Boolean;
begin
    Result := (GetObjectCount(Handle) = 0);
end;

{--------------------------------------------------------------------------}

procedure TOdbDriver.InternalPack( const Handle : TOdbHandle );
begin
end;

{--------------------------------------------------------------------------}

function TOdbDriver.GetDataAccessMethods : TDataAccessMethodSet;
begin
    Result := [damISAM];
end;

{--------------------------------------------------------------------------}

function TOdbDriver.CheckDuplicateInsert : Boolean;
begin
    Result := False;
end;

{--------------------------------------------------------------------------}

function TOdbDriver.CheckDuplicateUpdate : Boolean;
begin
    Result := False;
end;

{--------------------------------------------------------------------------}

procedure BuildCreateTableStatement( const Statement     : TStrings;
                                     const ORMapping     : TORMapping;
                                     const DatasetName   : String;
                                     const StringType    : String;
                                     const IntegerType   : String;
                                     const FloatType     : String;
                                     const BlobType      : String );
var
    Field        : TOdbField;
    FirstField   : Boolean;

    PrimaryIndex : TOdbIndex;
    NotNull      : String;

begin
    PrimaryIndex := ORMapping.GetPrimaryIndex(DatasetName);

    Statement.Clear;

    Statement.Add('create table ' + DatasetName);
    Statement.Add('(');

    FirstField := True;

    if ORMapping.Fields.GotoFirst then
    repeat
        if not FirstField then
            Statement.Add(',');

        Field := ORMapping.Fields.CurrentObj as TOdbField;

        { fields on the primary index must have the 'not null' attribute }
        if PrimaryIndex.HasField(Field.Name) then
            NotNull := 'not null'
        else
            NotNull := '';

        case Field.FieldType of
            oftString  :
            begin
                Statement.Add( Field.Name + ' ' + StringType + '(' + IntToStr(Field.Size-1) +
                               ') ' + NotNull  );
            end;

            oftInteger, oftAutoInc :
            begin
                Statement.Add( Field.Name + ' ' + IntegerType + ' ' +  NotNull  );
            end;

            oftFloat   :
            begin
                Statement.Add( Field.Name + ' ' + FloatType + ' ' + NotNull  );
            end;
        end;

        FirstField := False;

    until not ORMapping.Fields.GotoNext;

    Statement.Add(',');
    Statement.Add( 'Object ' + BlobType + ',' );

    Statement.Add( 'primary key (' + PrimaryIndex.FieldNamesAsString(',') + ')');

    Statement.Add(')');
end;

{--------------------------------------------------------------------------}

procedure BuildCreateIndexStatements(
                                          const Statements    : TStrings;
                                          const ObjectDataset : TObjectDataset );
var
    Index        : TOdbIndex;
    Line         : String;
begin

    Statements.Clear;

    { create the indices that are not primary }
    if ObjectDataset.Indices.GotoFirst then
    repeat
        Index := ObjectDataset.Indices.CurrentObj as TOdbIndex;

        if not (iaPrimary in Index.Attributes) then
        begin
            Line := 'create ';

            if (iaUnique in Index.Attributes) then
                Line := Line + 'unique ';

            if (iaDescending in Index.Attributes) then
                Line := Line + 'descending ';

            Line := Line + 'index ' + Index.Name + ' on ' + ObjectDataset.DatasetName +
                    ' (' + Index.FieldNamesAsString(',') + ')';

            Statements.Add(Line);

        end;
    until not ObjectDataset.Indices.GotoNext;
end;

{--------------------------------------------------------------------------}

procedure BuildSelectWhereClause( var   WhereClause   : String;
                                  const ORMapping     : TORMapping;
                                  const DatasetName   : String );
var
    PrimaryIndex   : TOdbIndex;
    Field          : TOdbField;
    FieldName      : String;
    FirstCondition : Boolean;
begin
    PrimaryIndex := ORMapping.GetPrimaryIndex(DatasetName);

    WhereClause    := 'where ';
    FirstCondition := True;

    if PrimaryIndex.FieldNames.GotoFirst then
    repeat
         FieldName := (PrimaryIndex.FieldNames.CurrentObj as TStringCombo).Str;

         Field     := GetFieldInList( FieldName, ORMapping.Fields );

         if not FirstCondition then
             WhereClause := WhereClause + ' and ';

         FirstCondition := False;

         WhereClause := WhereClause + FieldName + ' = ''' + Field.AsString + '''';

    until not PrimaryIndex.FieldNames.GotoNext;
end;

{--------------------------------------------------------------------------}

procedure BuildSelectObjectQuery( const SQL         : TStrings;
                                  const ORMapping   : TORMapping;
                                  const DatasetName : String );
var
    WhereClause    : String;
begin
    BuildSelectWhereClause(WhereClause, ORMapping, DatasetName);

    SQL.Clear;
    SQL.Add( 'select * from ' + DatasetName );
    SQL.Add( WhereClause );
end;

{--------------------------------------------------------------------------}

procedure BuildDeleteStatement( const SQL           : TStrings;
                                const ORMapping     : TORMapping;
                                const DatasetName   : String );
var
    PrimaryIndex   : TOdbIndex;
    Field          : TOdbField;
    WhereClause    : String;
    FieldName      : String;
    FirstCondition : Boolean;
begin
    PrimaryIndex := ORMapping.GetPrimaryIndex(DatasetName);

    WhereClause    := 'where ';
    FirstCondition := True;

    if PrimaryIndex.FieldNames.GotoFirst then
    repeat
         FieldName := (PrimaryIndex.FieldNames.CurrentObj as TStringCombo).Str;

         Field     := GetFieldInList( FieldName, ORMapping.Fields );

         if not FirstCondition then
             WhereClause := WhereClause + ' and ';

         FirstCondition := False;

         WhereClause := WhereClause + FieldName + ' = ''' + Field.AsString + '''';

    until not PrimaryIndex.FieldNames.GotoNext;

    SQL.Clear;
    SQL.Add( 'delete from ' + DatasetName );
    SQL.Add( WhereClause );
end;

{--------------------------------------------------------------------------}

procedure BuildInsertStatement( const SQL          : TStrings;
                                const Fields       : TUnOrderedList;
                                const DatasetName  : String );
var
    FieldNames  : String;
    FieldValues : String;
    Field       : TOdbField;

    function GetFieldValue( const Field : TOdbField ) : String;
    begin
        if Field.FieldType = oftString then
            Result := '''' + Field.AsString + ''''
        else
            Result := Field.AsString;
    end;

begin
    FieldNames  := '';
    FieldValues := '';

    if Fields.GotoFirst then
    repeat
        Field := Fields.CurrentObj as TOdbField;

        if FieldNames = '' then
            FieldNames := Field.Name
        else
            FieldNames := FieldNames + ', ' + Field.Name;

        if FieldValues = '' then
            FieldValues := GetFieldValue(Field)
        else
            FieldValues := FieldValues + ', ' + GetFieldValue(Field);

    until not Fields.GotoNext;

    SQL.Clear;
    SQL.Add( 'insert into ' + DatasetName );
    SQL.Add( '( ' + FieldNames +  ')' );
    SQL.Add( 'values (' + FieldValues + ')' );
end;

{--------------------------------------------------------------------------}

function CompareObjectDataset(const Lhs, Rhs : TObject) : TComparison; far;
begin
    Result := CompareInteger(LongInt(Lhs), LongInt(Rhs));
end;

{--------------------------------------------------------------------------}

constructor TObjectDatabase.Create(
                    const InOdbDriverClass : TOdbDriverClass;
                    const InDatabaseName   : String);
begin
    inherited Create;
    FDriver         := InOdbDriverClass.Create(InDatabaseName);
    FObjectDatasets := TUnOrderedList.Create(TObjectDataset, CompareObjectDataset);
    FObjectDatasets.Ownership := False;
end;

{--------------------------------------------------------------------------}

destructor TObjectDatabase.Destroy;
begin
    if Self.DatabaseOpened then
        Self.Close;

    FObjectDatasets.Free;
    FDriver.Free;
    inherited Destroy;
end;

{--------------------------------------------------------------------------}

procedure TObjectDatabase.StartTransaction;
begin
    FDriver.StartTransaction;
end;

{--------------------------------------------------------------------------}

procedure TObjectDatabase.Commit;
begin
    FDriver.Commit;
end;

{--------------------------------------------------------------------------}

procedure TObjectDatabase.Rollback;
begin
    FDriver.Rollback;
end;

{--------------------------------------------------------------------------}

procedure TObjectDatabase.Open(
                        const ServerName : String;
                        const UserName   : String;
                        const Password   : String
                      );
begin
    FDriver.OpenDatabase(ServerName, UserName, Password);
end;

{--------------------------------------------------------------------------}

procedure TObjectDatabase.Close;
var
    ObjectDataset : TObjectDataset;
begin
    { close all object files in the database. }
    if FObjectDatasets.GotoFirst then
    repeat
        ObjectDataset := (FObjectDatasets.CurrentObj as TObjectDataset);
        if ObjectDataset.Opened then
            ObjectDataset.Close;
    until not FObjectDatasets.GotoNext;

    FDriver.CloseDatabase;
end;

{--------------------------------------------------------------------------}

function TObjectDatabase.GetDatabaseName : String;
begin
    Result := FDriver.DatabaseName;
end;

{--------------------------------------------------------------------------}

function TObjectDatabase.GetDatabaseOpened : Boolean;
begin
    { MB Nov 29, 2000. The may be called when a driver object is destroyed.
      If the driver object is destroyed because an exception was raised in
      the constructor, FDriver may still be nil }
    if FDriver = nil then
        Result := False
    else
        Result := FDriver.DatabaseOpened;
end;

{--------------------------------------------------------------------------}

initialization

    OdbDriverRegistry := TUnOrderedList.Create(TStringCombo, CompareOdbDriverEntry);

finalization

    OdbDriverRegistry.Free;

end.
