{----------------------------------------------------------------------
|
| Library: Spider Object Database
|
| Module: OdbFF.Pas
|
| Description: Flash Filer object database driver class.
|
| History:    , 2001. Michel Brazeau, first version
|
| Notes: There are two problems with Flash Filer 2:
|        1. An error saying a file is already opened during a transaction
|        2. It is possible to delete a record that is locked.
|        These problems don't occur with Flash Filer 1. Currently,
|        transactions and object locking are not tested in the automated tests
|        with FF2. Do not use record locking or transactions with FF2.
|
|        Another issued workaround by having CheckDuplicateUpdate return True,
|        is that on a duplicate key update, the blob field is updated.
|
|---------------------------------------------------------------------}
unit OdbFF;

{$I Odb.Inc}

{$I FFDefine.Inc}

interface

uses
    {$ifdef __Testing}
    DriverTest, { for TOdbDriverTest }
    {$endif}

    Classes,  { for TMemoryStream }
    SysUtils, { for Exception, StrPas }
    ObjList,  { for TUnOrderedList }
    FFDb,     { for TFFDataset, TFFTable, TFFQuery }
    Odb,      { for TObjectFile, TOdbDriver }
    OdbIndex, { for TOdbIndex }
    OdbField; { for TOdbField }

type

{$ifdef FF2}
TOdbFFDataset = TFFDataset;
{$else}
TOdbFFDataset = TFFTable;
{$endif}

TFFOdbDriver = class(TOdbDriver)
protected

    FDatabase      : TFFDatabase;

public

    constructor Create(const InDatabaseName : String); override;
    destructor Destroy; override;

    procedure InternalStartTransaction; override;
    procedure InternalCommit; override;
    procedure InternalRollback; override;

    procedure InternalOpenDatabase(
                            const ServerName : String;
                            const UserName   : String;
                            const Password   : String
                          ); override;

    procedure InternalCloseDatabase; override;

    function InternalOpenDataset( const ObjectDataset : TObjectDataset ) : TOdbHandle; override;

    function InternalOpenExclusive( const ObjectFile : TObjectFile ) : TOdbHandle; override;

    procedure InternalCloseDataset( const Handle : TOdbHandle); override;

    procedure InternalDeleteDataset( const ObjectDataset : TObjectDataset ); override;

    procedure InternalCreateDataset( const ObjectDataset : TObjectDataset ); override;

    procedure InternalInsert(
                      const Handle  : TOdbHandle;
                      const Fields  : TUnOrderedList;
                      const Stream  : TMemoryStream ); override;

    procedure InternalUpdate(
                      const Handle     : TOdbHandle;
                      const Fields     : TUnOrderedList;
                      const Stream     : TMemoryStream ); override;

    procedure InternalDelete( const Handle : TOdbHandle ); override;

    function InternalCreateObjectUpdate : TOdbHandle; override;
    procedure InternalDestroyObjectUpdate( const Handle : TOdbHandle ); override;

    procedure InternalInsertSQL(
                      const Handle      : TOdbHandle;
                      const Stream      : TMemoryStream;
                      const ORMapping   : TORMapping;
                      const DatasetName : String ); override;

    procedure InternalUpdateSQL(
                      const Handle      : TOdbHandle;
                      const Stream      : TMemoryStream;
                      const ORMapping   : TORMapping;
                      const DatasetName : String ); override;

    procedure InternalDeleteSQL( const Handle      : TOdbHandle;
                                 const ORMapping   : TORMapping;
                                 const DatasetName : String ); override;

    function InternalGetIndexName( const Handle : TOdbHandle ) : String; override;

    procedure InternalSetIndexName(
                            const Handle    : TOdbHandle;
                            const IndexName : String ); override;

    procedure InternalGetCurrent(
                          const Handle  : TOdbHandle;
                          const Stream  : TMemoryStream ); override;

    function InternalLockCurrent( const Handle : TOdbHandle ) : Boolean; override;

    procedure InternalUnLockCurrent( const Handle : TOdbHandle ); override;

    function InternalFindNext( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindPrior( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindFirst( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindLast( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindExact(
                        const Handle : TOdbHandle;
                        const Fields : TUnOrderedList;
                        const Index  : TOdbIndex ) : Boolean; override;

    procedure InternalFindNearest(
                          const Handle : TOdbHandle;
                          const Fields : TUnOrderedList;
                          const Index  : TOdbIndex ); override;

    function InternalGetBookmark( const Handle : TOdbHandle ) : Pointer; override;
    procedure InternalGotoBookmark( const Handle   : TOdbHandle;
                                    const Bookmark : Pointer ); override;
    procedure InternalFreeBookmark( const Handle   : TOdbHandle;
                                    const Bookmark : Pointer ); override;

    function InternalGetObjectCount( const Handle : TOdbHandle ) : LongInt; override;

    function InternalIsEmpty( const Handle : TOdbHandle ) : Boolean; override;

    function GetSupportedFieldTypes : TOdbFieldTypes; override;

    function CheckDuplicateInsert : Boolean; override;
    function CheckDuplicateUpdate : Boolean; override;

    function GetFFDataset( const Handle : TOdbHandle ) : TOdbFFDataset;

    function GetDataAccessMethods : TDataAccessMethodSet; override;


    function AutoIncSupported              : Boolean; override;
    function IntegerIndexSupported         : Boolean; override;
    function FloatIndexSupported           : Boolean; override;
    function CaseInsensitiveIndexSupported : Boolean; override;
    function DescendingIndexSupported      : Boolean; override;
    function TransactionsSupported         : Boolean; override;
    function FileLockingSupported          : Boolean; override;
    function ObjectLockingSupported        : Boolean; override;
    function SingleIndexOnly               : Boolean; override;


protected
    function FFOpenFile( const Name       : String;
                         const ObjectFile : TObjectFile;
                         const Exclusive  : Boolean ) : TOdbHandle;

    {$ifdef FF2}
    function FFOpenQuery( const Name        : String;
                           const ObjectQuery : TObjectQuery ) : TOdbHandle;
    {$endif}

    procedure OpenTableForObjectUpdate( const Handle      : TOdbHandle;
                                        const ORMapping   : TORMapping;
                                        const DatasetName : String );

end;


{$ifdef __Testing}

type

TFFOdbDriverTest = class( TOdbDriverTest )
protected
     procedure GetParams( var DriverClass                   : TOdbDriverClass;
                          var DatabaseName                  : String;
                          var ServerName                    : String;
                          var UserName                      : String;
                          var Password                      : String
                        ); override;
end;

{$endif}

{--------------------------------------------------------------------}

implementation

uses
    Db,       { for ftString, ftInteger, ftFloat }
    {$ifdef FF2}
    FFLLBase, { for ffltWriteLock }
    {$else}
    FFClEng,  { for VEng }
    Bde,      { for dbiWriteLock }
    {$endif}
    ObjPerst,
    ObjBuckt, { for TStringCombo }
    OdbMisc;  { for RaiseOdbException }

const

CObjectFieldName = 'Object';

{--------------------------------------------------------------------}

var

{ last database driver instance that was created. }
LastInstanceNo   : LongInt;

{--------------------------------------------------------------------}

constructor TFFOdbDriver.Create(const InDatabaseName : String);
begin
    FDatabase := TFFDatabase.Create(nil);

    Inc(LastInstanceNo);

    FDatabase.AliasName    := InDatabaseName;
    FDatabase.DatabaseName := 'Odb' + IntToStr(LastInstanceNo);

    inherited Create('Odb');
end;

{--------------------------------------------------------------------}

destructor TFFOdbDriver.Destroy;
begin
    { invoke inherited first, to close the database if it is opened before
      invoking FDatabase.Free }
    inherited Destroy;

    FDatabase.Free;
end;

{--------------------------------------------------------------------}

procedure TFFOdbDriver.InternalStartTransaction;
begin
    FDatabase.StartTransaction;
end;

{--------------------------------------------------------------------}

procedure TFFOdbDriver.InternalCommit;
begin
    FDatabase.Commit;
end;

{--------------------------------------------------------------------}

procedure TFFOdbDriver.InternalRollback;
begin
    FDatabase.Rollback;
end;

{--------------------------------------------------------------------}

procedure TFFOdbDriver.InternalOpenDatabase(
                        const ServerName : String;
                        const UserName   : String;
                        const Password   : String
                      );
begin
    FDatabase.Open;
end;

{--------------------------------------------------------------------}

procedure TFFOdbDriver.InternalCloseDatabase;
begin
    FDatabase.Close;
end;

{--------------------------------------------------------------------}

function TFFOdbDriver.FFOpenFile( const Name       : String;
                                          const ObjectFile : TObjectFile;
                                          const Exclusive  : Boolean ) : TOdbHandle;
var
    Table : TFFTable;

begin
    Table := TFFTable.Create(nil);
    try
        Table.DatabaseName := FDatabase.DatabaseName;
        Table.TableName    := Name;
        Table.Exclusive    := Exclusive;
        Table.Open;

        Result := NewHandle(Table);
    except
        Table.Free;
        raise;
    end;
end;

{--------------------------------------------------------------------}

{$ifdef FF2}
function TFFOdbDriver.FFOpenQuery(
                           const Name        : String;
                           const ObjectQuery : TObjectQuery ) : TOdbHandle;
var
    Query     : TFFQuery;

begin
    Query     := nil;
    try
        Query     := TFFQuery.Create(nil);

        Query.DatabaseName := FDatabase.DatabaseName;
        Query.SQL.Assign( ObjectQuery.Select );
        Query.Open;

        Result := NewHandle(Query);
    except
        Query.Free;
        raise;
    end;
end;
{$endif}

{--------------------------------------------------------------------}

function TFFOdbDriver.InternalOpenDataset(
                       const ObjectDataset : TObjectDataset ) : TOdbHandle;
begin
    {$ifdef FF2}
    if ObjectDataset is TObjectFile then
        Result := FFOpenFile( ObjectDataset.DatasetName,
                              TObjectFile(ObjectDataset), False)
    else if ObjectDataset is TObjectQuery then
    begin
        if TObjectQuery(ObjectDataset).Select.Count = 0 then
            RaiseOdbException( EOdbSelectStatementNotSpecified,
                               ObjectDataset.DatasetName );

        Result := FFOpenQuery( ObjectDataset.DatasetName, TObjectQuery(ObjectDataset) );
    end
    else
    begin
        Result := 0; { prevent compiler warning }
        RaiseOdbException(EOdbError, 'Internal error ' + ObjectDataset.DatasetName);
    end;
    {$else}
    Result := FFOpenFile(Name, TObjectFile(ObjectDataset), False);
    {$endif}
end;
{--------------------------------------------------------------------}

function TFFOdbDriver.InternalOpenExclusive(
                                  const ObjectFile : TObjectFile ) : TOdbHandle;
begin
    Result := FFOpenFile(ObjectFile.DatasetName, ObjectFile, True);
end;

{--------------------------------------------------------------------}

procedure TFFOdbDriver.InternalCloseDataset( const Handle : TOdbHandle);
var
    Dataset : TOdbFFDataset;
begin
    Dataset := GetFFDataset(Handle);
    Dataset.Close;

    ReleaseHandle(Handle);
end;

{--------------------------------------------------------------------}

procedure TFFOdbDriver.InternalDeleteDataset( const ObjectDataset : TObjectDataset );
var
    Table : TFFTable;
begin
    Table := TFFTable.Create(nil);
    try
        Table.DatabaseName := FDatabase.DatabaseName;
        Table.TableName    := ObjectDataset.DatasetName;
        Table.DeleteTable;
    finally
        Table.Free;
    end;
end;

{--------------------------------------------------------------------}

function TFFOdbDriver.InternalCreateObjectUpdate : TOdbHandle;
var
    Dataset : TFFTable;

begin
    Dataset := nil;
    try
        { MB Mar 22, 2001. Flash filer does not support TQuery based
          updating }
        Dataset := TFFTable.Create(nil);

        Dataset.DatabaseName := FDatabase.DatabaseName;

        Result := NewHandle(Dataset);
    except
        Dataset.Free;
        raise;
    end;
end;

{--------------------------------------------------------------------}

procedure TFFOdbDriver.InternalDestroyObjectUpdate( const Handle : TOdbHandle );
begin
    ReleaseHandle(Handle);
end;

{--------------------------------------------------------------------}

{ add Fields to Table }
procedure AddFields( const Fields  : TUnOrderedList;
                     const Table   : TFFTable );
var
    Field   : TOdbField;
begin
    if Fields.GotoFirst then
    repeat
        Field := Fields.CurrentObj as TOdbField;

        case Field.FieldType of
            oftString  : Table.FieldDefs.Add(
                                 Field.Name,
                                 Db.ftString, Field.Size, False );
            oftInteger : Table.FieldDefs.Add(
                                 Field.Name,
                                 Db.ftInteger, 0, False );
            oftFloat   : Table.FieldDefs.Add(
                                 Field.Name,
                                 Db.ftFloat, 0, False );
            oftAutoInc : Table.FieldDefs.Add(
                                 Field.Name,
                                 Db.ftAutoInc, 0, False );

        end;
    until not Fields.GotoNext;

    Table.FieldDefs.Add(CObjectFieldName, ftBlob, 0, False);
end;

{--------------------------------------------------------------------}

{ convert Spider Odb index attributes to VCL index options }
function IndexAttributesToOptions(
            const Attributes : TOdbIndexAttributes ) : TIndexOptions;
begin
    Result := [];

    if iaUnique in Attributes then
        Result := Result + [ixUnique];

    if iaCaseInsensitive in Attributes then
        Result := Result + [ixCaseInsensitive];

    if iaDescending in Attributes then
        Result := Result + [ixDescending];

    if iaPrimary in Attributes then
        Result := Result + [ixPrimary];
end;

{--------------------------------------------------------------------}

procedure AddIndices( const Indices          : TUnOrderedList;
                      const Table            : TFFTable );
var
    Index        : TOdbIndex;
    IndexOptions : TIndexOptions;
    FieldNames   : String;
    IndexName    : String;

begin
    if Indices.GotoFirst then
    repeat
        Index        := Indices.CurrentObj as TOdbIndex;
        FieldNames   := Index.FieldNamesAsString(';');
        IndexOptions := IndexAttributesToOptions(Index.Attributes);

        IndexName := Index.Name;

        if ixPrimary in IndexOptions then
            IndexOptions := IndexOptions - [ixPrimary] + [ixUnique];

        Table.IndexDefs.Add( IndexName,
                             FieldNames,
                             IndexOptions );

    until not Indices.GotoNext;
end;

{--------------------------------------------------------------------}

procedure TFFOdbDriver.InternalCreateDataset(
                          const ObjectDataset : TObjectDataset
                        );
var
    Table : TFFTable;

begin
    Table := TFFTable.Create(nil);
    try
        Table.DatabaseName := FDatabase.DatabaseName;
        Table.TableName    := ObjectDataset.DatasetName;

        Table.FieldDefs.Clear;
        Table.IndexDefs.Clear;

        AddFields(ObjectDataset.Fields, Table);
        AddIndices(ObjectDataset.Indices, Table);
        Table.CreateTable;

    finally
        Table.Free;
    end;
end;

{--------------------------------------------------------------------}

{ set the field values in Table given a list of TOdbField's }
procedure SetDatasetFieldValues( const Fields  : TUnOrderedList;
                                 const Dataset : TOdbFFDataset );
var
    Field : TOdbField;
begin
    if Fields.GotoFirst then
    repeat
        Field := Fields.CurrentObj as TOdbField;

        case Field.FieldType of
            oftString  : Dataset.FieldByName(Field.Name).AsString  := Field.AsString;
            oftInteger : Dataset.FieldByName(Field.Name).AsInteger := Field.AsInteger;
            oftFloat   : Dataset.FieldByName(Field.Name).AsFloat   := Field.AsFloat;
        end;
    until not Fields.GotoNext;
end;

{--------------------------------------------------------------------}

{ set the field values in Table for the fields belonging to Index }
procedure SetTableFieldValuesOnIndex( const Fields  : TUnOrderedList;
                                      const Index   : TOdbIndex;
                                      const Table   : TFFTable );
var
    Field       : TOdbField;
    StringCombo : TStringCombo;
begin
    if Index.FieldNames.GotoFirst then
    repeat
        StringCombo := Index.FieldNames.CurrentObj as TStringCombo;

        Field := GetFieldInList( StringCombo.Str, Fields );

        case Field.FieldType of
            oftString  : Table.FieldByName(Field.Name).AsString := Field.AsString;
            oftInteger,
            oftAutoInc : Table.FieldByName(Field.Name).AsInteger := Field.AsInteger;
            oftFloat   : Table.FieldByName(Field.Name).AsFloat := Field.AsFloat;
        end;
    until not Index.FieldNames.GotoNext;
end;

{--------------------------------------------------------------------}

{ Set the value of the autoincrement fields in Fields from the values
  contained in Table. }
procedure SetOdbAutoIncFieldValues( const Fields  : TUnOrderedList;
                                    const Table   : TFFTable );
var
    Field : TOdbField;
begin
    if Fields.GotoFirst then
    repeat
        Field := Fields.CurrentObj as TOdbField;

        case Field.FieldType of
            oftAutoInc : Field.AsInteger := Table.FieldByName(Field.Name).AsInteger;
        end;
    until not Fields.GotoNext;
end;

{--------------------------------------------------------------------}

procedure InsertInDataset( const Dataset : TOdbFFDataset;
                           const Fields  : TUnOrderedList;
                           const Stream  : TMemoryStream );
var
    Field   : TBlobField;
begin
    Dataset.Insert;
    try
        SetDatasetFieldValues(Fields, Dataset);

        Field := (Dataset.FieldByName(CObjectFieldName) as TBlobField);

        Field.LoadFromStream(Stream);

        Dataset.Post;

        { Update any autoincrement field in Fields. }
        if Dataset is TFFTable then
             SetOdbAutoIncFieldValues(Fields, TFFTable(Dataset));
    except
        Dataset.Cancel;
        raise;
    end;
end;

{--------------------------------------------------------------------}

procedure TFFOdbDriver.InternalInsert(
                  const Handle  : TOdbHandle;
                  const Fields  : TUnOrderedList;
                  const Stream  : TMemoryStream );

var
    Dataset : TOdbFFDataset;
begin
    Dataset := GetFFDataset(Handle);

    InsertInDataset(Dataset, Fields, Stream);
end;

{--------------------------------------------------------------------}

procedure UpdateCurrentInDataset( const Dataset : TOdbFFDataset;
                                  const Fields  : TUnOrderedList;
                                  const Stream  : TMemoryStream );
begin
     try
        SetDatasetFieldValues(Fields, Dataset);
        (Dataset.FieldByName(CObjectFieldName) as TBlobField).LoadFromStream(Stream);
        Dataset.Post;
    except
        Dataset.Cancel;
        raise;
    end;
end;

{--------------------------------------------------------------------}

procedure TFFOdbDriver.InternalUpdate(
                  const Handle     : TOdbHandle;
                  const Fields     : TUnOrderedList;
                  const Stream     : TMemoryStream );

var
    Dataset      : TOdbFFDataset;
begin
    Dataset      := GetFFDataset(Handle);

    { MB May 21, 1998. Dataset.Edit will attempt to lock the record before
      updating it. If the record has already been locked by the current session,
      it is acceptable to perform an update. A try,except block is used to
      attempt to unlock the object. If the object is locked by another session,
      the second call to Dataset.Edit will raise the exception. }
    try
        Dataset.Edit;
    except
        on Exception do
        begin
            InternalUnLockCurrent(Handle);
            Dataset.Edit;
        end;
    end;

    UpdateCurrentInDataset(Dataset, Fields, Stream);
end;

{--------------------------------------------------------------------}

procedure TFFOdbDriver.InternalDelete( const Handle : TOdbHandle );
var
    Dataset       : TOdbFFDataset;
begin
    Dataset       := GetFFDataset(Handle);

    Dataset.Delete
end;

{--------------------------------------------------------------------}

procedure TFFOdbDriver.OpenTableForObjectUpdate(
                  const Handle      : TOdbHandle;
                  const ORMapping   : TORMapping;
                  const DatasetName : String );

var
    Table       : TFFTable;
begin
    Table := GetFFDataset(Handle) as TFFTable;
    if not Table.Active then
    begin
        Table.TableName        := DatasetName;
        InternalSetIndexName(Handle, ORMapping.GetPrimaryIndex(DatasetName).Name);
        Table.Open;
    end;
end;

{--------------------------------------------------------------------}

procedure TFFOdbDriver.InternalInsertSQL(
                  const Handle      : TOdbHandle;
                  const Stream      : TMemoryStream;
                  const ORMapping   : TORMapping;
                  const DatasetName : String );
begin
    OpenTableForObjectUpdate( Handle, ORMapping, DatasetName );

    InternalInsert(Handle, ORMapping.Fields, Stream );
end;

{--------------------------------------------------------------------}

procedure TFFOdbDriver.InternalUpdateSQL(
                      const Handle      : TOdbHandle;
                      const Stream      : TMemoryStream;
                      const ORMapping   : TORMapping;
                      const DatasetName : String );
begin
    OpenTableForObjectUpdate( Handle, ORMapping, DatasetName );

    if not InternalFindExact( Handle, ORMapping.Fields, ORMapping.GetPrimaryIndex(DatasetName) ) then
        RaiseOdbException( EOdbObjectNotFound, 'TBdeOdbDriver.InternalUpdateSQL' );

    InternalUpdate(Handle, ORMapping.Fields, Stream );
end;

{--------------------------------------------------------------------}

procedure TFFOdbDriver.InternalDeleteSQL(
                             const Handle      : TOdbHandle;
                             const ORMapping   : TORMapping;
                             const DatasetName : String );
begin
    OpenTableForObjectUpdate( Handle, ORMapping, DatasetName );

    if not InternalFindExact( Handle, ORMapping.Fields, ORMapping.GetPrimaryIndex(DatasetName) ) then
        RaiseOdbException( EOdbObjectNotFound, 'TBdeOdbDriver.InternalDeleteSQL' );

    InternalDelete(Handle);
end;

{--------------------------------------------------------------------}

function TFFOdbDriver.InternalGetIndexName( const Handle : TOdbHandle ) : String;
var
    Table : TFFTable;
begin
    Table  := GetFFDataset(Handle) as TFFTable;
    Result := Table.IndexName;
end;

{--------------------------------------------------------------------}

procedure TFFOdbDriver.InternalSetIndexName(
                      const Handle    : TOdbHandle;
                      const IndexName : String );
var
    Table : TFFTable;
begin
    Table := GetFFDataset(Handle) as TFFTable;

    Table.IndexName := IndexName;
end;

{--------------------------------------------------------------------}

procedure TFFOdbDriver.InternalGetCurrent(
                      const Handle  : TOdbHandle;
                      const Stream  : TMemoryStream );
var
    Dataset : TOdbFFDataset;
    Field   : TBlobField;
begin
    Dataset := GetFFDataset(Handle) as TOdbFFDataset;

    Field := (Dataset.FieldByName(CObjectFieldName) as TBlobField);

    Field.SaveToStream(Stream);
end;

{--------------------------------------------------------------------}

function TFFOdbDriver.InternalLockCurrent( const Handle : TOdbHandle ) : Boolean;
var
    Table    : TFFTable;
begin
    Table := GetFFDataset(Handle) as TFFTable;

    Table.UpdateCursorPos;

    {$ifdef FF2}
    Result := (Table.ServerEngine.RecordGet( Table.CursorID,
                                  ffltWriteLock,
                                  Pointer(Table.ActiveBuffer)) = 0);
    {$else}
    Result := (FFClEng.VEng.GetRecord(Table.Handle, dbiWriteLock, Table.ActiveBuffer, nil) = 0);
    {$endif}
end;


{--------------------------------------------------------------------}

procedure TFFOdbDriver.InternalUnLockCurrent( const Handle : TOdbHandle );
var
    Table : TFFTable;
begin
    Table := GetFFDataset(Handle) as TFFTable;

    Table.UpdateCursorPos;

    {$ifdef FF2}
    Table.ServerEngine.RecordRelLock( Table.CursorID, False);
    {$else}
    FFClEng.VEng.RelRecordLock(Table.Handle, False);
    {$endif}
end;

{--------------------------------------------------------------------}

function TFFOdbDriver.InternalFindNext( const Handle : TOdbHandle ) : Boolean;
var
    Dataset : TOdbFFDataset;
begin
    Dataset := GetFFDataset(Handle);
    Dataset.Next;
    Result := not Dataset.EOF;
end;

{--------------------------------------------------------------------}

function TFFOdbDriver.InternalFindPrior( const Handle : TOdbHandle ) : Boolean;
var
    Dataset : TOdbFFDataset;
begin
    Dataset := GetFFDataset(Handle);
    Dataset.Prior;
    Result := not Dataset.BOF;
end;

{--------------------------------------------------------------------}

function TFFOdbDriver.InternalFindFirst( const Handle : TOdbHandle ) : Boolean;
var
    Dataset : TOdbFFDataset;
begin
    Dataset := GetFFDataset(Handle);
    Dataset.First;
    Result := not (Dataset.EOF and Dataset.BOF);
end;

{--------------------------------------------------------------------}

function TFFOdbDriver.InternalFindLast( const Handle : TOdbHandle ) : Boolean;
var
    Dataset : TOdbFFDataset;
begin
    Dataset := GetFFDataset(Handle);
    Dataset.Last;
    Result := not (Dataset.EOF and Dataset.BOF);
end;

{--------------------------------------------------------------------}

function TFFOdbDriver.InternalFindExact(
                    const Handle : TOdbHandle;
                    const Fields : TUnOrderedList;
                    const Index  : TOdbIndex ) : Boolean;
var
    Table : TFFTable;
begin
    Table := GetFFDataset(Handle) as TFFTable;
    Table.SetKey;
    SetTableFieldValuesOnIndex(Fields, Index, Table);
    Result := Table.GotoKey;
end;

{--------------------------------------------------------------------}

procedure TFFOdbDriver.InternalFindNearest(
                      const Handle : TOdbHandle;
                      const Fields : TUnOrderedList;
                      const Index  : TOdbIndex );
var
    Table : TFFTable;
begin
    Table := GetFFDataset(Handle) as TFFTable;
    Table.SetKey;
    SetTableFieldValuesOnIndex(Fields, Index, Table);
    Table.GotoNearest;
end;

{--------------------------------------------------------------------}

function TFFOdbDriver.InternalGetBookmark( const Handle : TOdbHandle ) : Pointer;
var
    Table : TFFTable;
begin
    Table := GetFFDataset(Handle) as TFFTable;

    Result := Table.GetBookmark;
end;

{--------------------------------------------------------------------}

procedure TFFOdbDriver.InternalGotoBookmark( const Handle   : TOdbHandle;
                                              const Bookmark : Pointer );
var
    Table : TFFTable;
begin
    Table := GetFFDataset(Handle) as TFFTable;

    Table.GotoBookmark(Bookmark);
end;

{--------------------------------------------------------------------}

procedure TFFOdbDriver.InternalFreeBookmark( const Handle   : TOdbHandle;
                                              const Bookmark : Pointer );
var
    Table : TFFTable;
begin
    Table := GetFFDataset(Handle) as TFFTable;

    Table.FreeBookmark(Bookmark);
end;


{--------------------------------------------------------------------}

function TFFOdbDriver.InternalGetObjectCount( const Handle : TOdbHandle ) : LongInt;
begin
    Result := (GetFFDataset(Handle) as TFFTable).RecordCount;
end;

{--------------------------------------------------------------------}

function TFFOdbDriver.InternalIsEmpty( const Handle : TOdbHandle ) : Boolean;
var
    Dataset : TOdbFFDataset;
begin
    Dataset := GetFFDataset(Handle);
    Result  := Dataset.BOF and Dataset.EOF;
end;

{--------------------------------------------------------------------}

function TFFOdbDriver.GetFFDataset( const Handle : TOdbHandle ) : TOdbFFDataset;
begin
    Result := GetHandleDataObj(Handle) as TOdbFFDataset;
end;

{--------------------------------------------------------------------}

function TFFOdbDriver.GetDataAccessMethods : TDataAccessMethodSet;
begin
    Result := [damISAM, damSQL];
end;

{--------------------------------------------------------------------}

function TFFOdbDriver.GetSupportedFieldTypes : TOdbFieldTypes;
begin
    Result := [oftString, oftInteger, oftFloat, oftAutoInc];
end;

{--------------------------------------------------------------------}

function TFFOdbDriver.CheckDuplicateInsert : Boolean;
begin
    Result := False;
end;

{--------------------------------------------------------------------}

function TFFOdbDriver.CheckDuplicateUpdate : Boolean;
begin
    { MB Mar 11, 2000. Workaround a bug in FF engine. On a duplicate
      key update, an exception is raised as expected, but the record is
      somehow updated. This was discovered with the automated tests

      See the comment in DriverTest.TestDuplicateUpdateObjectFile }

    Result := True;
end;

{--------------------------------------------------------------------------}

function TFFOdbDriver.AutoIncSupported              : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TFFOdbDriver.IntegerIndexSupported         : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TFFOdbDriver.FloatIndexSupported           : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TFFOdbDriver.CaseInsensitiveIndexSupported : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TFFOdbDriver.DescendingIndexSupported      : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TFFOdbDriver.TransactionsSupported         : Boolean;
begin
    { see notes in unit header }
    {$ifdef FF2}
    Result := False;
    {$else}
    Result := True;
    {$endif}
end;

{--------------------------------------------------------------------------}

function TFFOdbDriver.FileLockingSupported          : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TFFOdbDriver.ObjectLockingSupported        : Boolean;
begin
    { see notes in unit header }
    {$ifdef FF2}
    Result := False;
    {$else}
    Result := True;
    {$endif}
end;

{--------------------------------------------------------------------------}

function TFFOdbDriver.SingleIndexOnly               : Boolean;
begin
    Result := False;
end;

{--------------------------------------------------------------------}

{$ifdef __Testing}

procedure TFFOdbDriverTest.GetParams(
                    var DriverClass                   : TOdbDriverClass;
                    var DatabaseName                  : String;
                    var ServerName                    : String;
                    var UserName                      : String;
                    var Password                      : String
                  );
begin
    DriverClass  := TFFOdbDriver;
    DatabaseName := 'OdbTest';
    ServerName   := '';
    UserName     := '';
    Password     := '';
end;

{$endif}

{--------------------------------------------------------------------}


initialization
    LastInstanceNo := 0;

    RegisterOdbDriver(TFFOdbDriver, 'FlashFiler');
end.
