{---------------------------------------------------------------------------
|
|  Module: OdbField.Pas
|
|  Description: Database field abstraction classes.
|
|  History: Dec 31, 1997 Michel Brazeau, first version.
|           Sep 12, 1998 Change the type of InSize to Word to allow
|                        a size of 256 (a string of 255 characters).
|           Feb   , 2001. MB Create unit tests for all routines
|
|---------------------------------------------------------------------------}
unit OdbField;

{$I Odb.Inc}

interface

uses
    {$ifdef __Testing}
    AutoTest, { for TSpiderOdbTestcase }
    {$endif}

    SysUtils, { for Exception, PString }
    ObjCmp,   { for TComparison, CompareStringNoCase }
    ObjList;  { for TUnOrderedList }


{--------------------------------------------------------------------------}

{ field related types }

type

{ don't change ordering as ClassGen wizard relies on this ordering. }
TOdbFieldType = ( oftString,
                  oftFloat,
                  oftInteger,
                  oftAutoInc
                );
TOdbFieldTypes = set of TOdbFieldType;

TOdbField = class(TObject)
protected

    { MB Oct 25, 2000. Change from PString to AnsiString }
    FName    : AnsiString;

    FType    : TOdbFieldType;
    FSize    : Word;

    FDataPtr : Pointer;

    function GetName : String;
    procedure SetName( const InName : String );

    function GetAsString : String; virtual;
    procedure SetAsString(const Value : String); virtual;

    function GetAsFloat : Double; virtual;
    procedure SetAsFloat(const Value : Double); virtual;

    function GetAsInteger : LongInt; virtual;
    procedure SetAsInteger(const Value : LongInt); virtual;

    { MB Oct 05, 2001. Needed since FieldType is now "writable" }
    procedure SetFieldType( const InFieldType : TOdbFieldType );
    procedure SetSize( const InSize : Word );

public

    { MB Oct 24, 2000. General purpose "tag" which may be used by a driver
      to store information related to the field. It is put directly in the
      public section, so its address may be pass as a parameter to
      driver related functions. For example, the OdbOdbc.Pas driver uses
      this to store the length field passed to SQLBindParameter }
    Tag  : LongInt;

    { InSize is only used for string fields. String fields are stored
      using length byte strings, and InSize must include the length byte
      itself. }
    constructor Create( const InName       : String;
                        const InType       : TOdbFieldType;
                        const InSize       : Word
                      );

    destructor Destroy; override;

    { returns the address of the first character of the string when the
      FieldType is oftString. If the FieldType is not oftString an
      exception is raised. }
    function GetStrAddress : Pointer;

    { AsXXXX methods may raise EOdbInvalidFieldType }

    property AsString : String read GetAsString
                               write SetAsString;

    property AsInteger : LongInt read GetAsInteger
                                write SetAsInteger;

    property AsFloat : Double read GetAsFloat
                                write SetAsFloat;


    { MB Feb 13, 2001. Make FieldType and Size readonly. They must be set
      by the constructor as memory is allocated in the constructor.
      MB Oct 05, 2001. A beta tester of version reported requiring FieldType
      writable. See also SetFieldType }
    property FieldType : TOdbFieldType read FType write SetFieldType;
    property Name      : String read GetName write SetName;
    property Size      : Word read FSize write SetSize;

    property DataPtr   : Pointer read FDataPtr;

private
    { raises EOdbInvalidFieldType if InFieldType is not compatible with
      Self.FieldType }
    procedure CheckType( const InFieldType : TOdbFieldType );

end;

function CompareOdbField(const Lhs, Rhs : TObject) : TComparison; far;

{ returns a field in a field list. May raise EOdbFieldDoesNotExist }
function GetFieldInList( const FieldName : String;
                         const FieldList : TUnOrderedList ) : TOdbField;

{ returns True if the value if Field1 is equal to the value of Field2. May
  raise EOdbFieldsNotComparable. }
function FieldsEqual( const Field1 : TOdbField;
                      const Field2 : TOdbField ) : Boolean;

{ Returns True if FieldList has at least one autoincrement field. }
function FieldListHasAutoInc( const FieldList : TUnOrderedList ) : Boolean;


{$ifdef __Testing}

type

TOdbFieldUnitTest = class( TSpiderOdbTestCase )
published
     procedure TestOdbField;
     procedure TestCompareOdbField;
     procedure TestGetFieldInList;
     procedure TestFieldsEqual;
     procedure TestFieldListHasAutoInc;
end;

{$endif}

{--------------------------------------------------------------------}

implementation

uses
    ObjPerst, { for TObjectReader, TObjectWriter }
    OdbMisc,  { for RaiseOdbException, TruncateStr }
    Odb;      { for EOdbFieldDoesNotExist }


type
    Str255 = String[255];

{--------------------------------------------------------------------------}

function GetFieldInList( const FieldName : String;
                         const FieldList : TUnOrderedList ) : TOdbField;
var
    Key : TOdbField;
begin
    Result := nil; { prevent compiler warning }

    Key := TOdbField.Create( FieldName, oftInteger, 0 );
    try
        if FieldList.Search(Key, 1) then
            Result := (FieldList.CurrentObj as TOdbField)
        else
            RaiseOdbException( EOdbFieldDoesNotExist, FieldName );
    finally
        Key.Free;
    end;
end;

{--------------------------------------------------------------------------}

function FieldListHasAutoInc( const FieldList : TUnOrderedList ) : Boolean;
var
    Field : TOdbField;
begin
    Result := False;

    if FieldList.GotoFirst then
    repeat
        Field := FieldList.CurrentObj as TOdbField;

        if Field.FieldType = oftAutoInc then
        begin
            Result := True;
            break;
        end;
    until not FieldList.GotoNext;
end;

{--------------------------------------------------------------------------}

{ returns True if two field types are comparable }
function FieldTypesComparable( const Type1 : TOdbFieldType;
                               const Type2 : TOdbFieldType ) : Boolean;
begin
    Result := True; { assume success }

    if Type1 = Type2 then
        Exit;

    Result := (
               (Type1 = oftInteger) and
               (Type2 = oftAutoInc)
              )
              or
              (
               (Type2 = oftInteger) and
               (Type1 = oftAutoInc)
              );
end;

{--------------------------------------------------------------------------}

function FieldsEqual( const Field1 : TOdbField;
                      const Field2 : TOdbField ) : Boolean;
begin
    Result := False; { prevent compiler warning }

    if not FieldTypesComparable(Field1.FieldType, Field2.FieldType) then
        RaiseOdbException( EOdbFieldsNotComparable, '' );

    case Field1.FieldType of
        oftString  : Result := (Field1.AsString = Field2.AsString);
        oftInteger,
        oftAutoInc : Result := (Field1.AsInteger = Field2.AsInteger);
        oftFloat   : Result := (Field1.AsFloat = Field2.AsFloat);
        else
            RaiseOdbException( EOdbFieldTypeNotSupported, '' );
    end;
end;

{--------------------------------------------------------------------------}

function CompareOdbField(const Lhs, Rhs : TObject) : TComparison;
var
    Left  : TOdbField absolute Lhs;
    Right : TOdbField absolute Rhs;
begin
    Result := CompareStringNoCase(Left.Name, Right.Name);
end;

{--------------------------------------------------------------------------}

procedure ReadOdbField( const Obj     : TObject;
                        const Reader  : TObjectReader );
var
    Field : TOdbField;
begin
    Field := Obj as TOdbField;

    Field.FName := Reader.ReadString;
    Field.FType := TOdbFieldType(Reader.ReadInteger);
    Field.FSize := Reader.ReadInteger;

    GetMem(Field.FDataPtr, Field.FSize);
    Reader.Read(Field.FDataPtr^, Field.FSize);
end;

{--------------------------------------------------------------------------}

procedure WriteOdbField( const Obj     : TObject;
                         const Writer  : TObjectWriter );
var
    Field : TOdbField;
begin
    Field := Obj as TOdbField;

    Writer.WriteString(Field.FName);
    Writer.WriteInteger(Ord(Field.FType));
    Writer.WriteInteger(Field.FSize);

    Writer.Write(Field.FDataPtr^, Field.FSize);
end;

{--------------------------------------------------------------------------}

constructor TOdbField.Create(
                    const InName       : String;
                    const InType       : TOdbFieldType;
                    const InSize       : Word
                    );
begin
    inherited Create;

    FName     := InName;

    { MB Oct 05, 2001 }
    FieldType := InType;
    Size      := InSize;
end;

{-------------------------------------------------------------------------}

destructor TOdbField.Destroy;
begin
    FreeMem(FDataPtr);

    inherited Destroy;
end;

{-------------------------------------------------------------------------}

procedure TOdbField.SetFieldType( const InFieldType : TOdbFieldType );
begin
    case InFieldType of

        oftString  :
        begin
            { Use a default string size. The size property may then be used
              to change it }
            FSize := 41;
        end;

        oftInteger,
        oftAutoInc : FSize := SizeOf(LongInt);

        oftFloat   : FSize := SizeOf(Double);
    end;

    FType := InFieldType;

    GetMem(FDataPtr, FSize);
end;
{-------------------------------------------------------------------------}

procedure TOdbField.SetSize( const InSize : Word );
begin
    { SetSize only applies with String type fields }
    if FType = oftString then
    begin
        FreeMem(FDataPtr);
        FSize := InSize;
        GetMem(FDataPtr, FSize);
    end;
end;

{-------------------------------------------------------------------------}

function TOdbField.GetName : String;
begin
    Result := FName;
end;

{-------------------------------------------------------------------------}

procedure TOdbField.SetName( const InName : String );
begin
    FName := InName;
end;

{-------------------------------------------------------------------------}

procedure TOdbField.CheckType( const InFieldType : TOdbFieldType );
var
    Incompatible : Boolean;
begin
    Incompatible := True;
    case Self.FieldType of
        oftString,
        oftFloat   : Incompatible := (InFieldType <> Self.FieldType);

        oftInteger,
        oftAutoInc : Incompatible := (InFieldType = oftString) or
                                     (InFieldType = oftFloat);
    end; { case }

    if Incompatible then
        RaiseOdbException( EOdbInvalidFieldType, '' );
end;

{-------------------------------------------------------------------------}

function TOdbField.GetStrAddress : Pointer;
begin
    if FieldType <> oftString then
        RaiseOdbException( EOdbInvalidFieldType, '' );

    Result := @Str255(FDataPtr^)[1];
end;

{-------------------------------------------------------------------------}

function TOdbField.GetAsString : String;
begin
    { MB Sep 27, 2000. All field types allow returning AsString. This
      allows creating SQL statements more easily }

    case FType of

        oftString  : Result := Str255(FDataPtr^);

        oftFloat   : Result := FloatToStrF( Double(FDataPtr^),
                                            ffFixed, 18, 15 );


        oftInteger,
        oftAutoInc : Result := IntToStr(LongInt(FDataPtr^));

        else
            RaiseOdbException( EOdbError, 'Internal error, TOdbField.GetAsString');

    end;

end;

{-------------------------------------------------------------------------}

procedure TOdbField.SetAsString(const Value : String);
begin
    CheckType(oftString);

    Str255(FDataPtr^) := TruncateStr(Value, Size-1);
end;

{-------------------------------------------------------------------------}

function TOdbField.GetAsFloat : Double;
begin
    CheckType(oftFloat);

    Result := Double(FDataPtr^);
end;

{-------------------------------------------------------------------------}

procedure TOdbField.SetAsFloat(const Value : Double);
begin
    CheckType(oftFloat);

    Double(FDataPtr^) := Value;
end;

{-------------------------------------------------------------------------}

function TOdbField.GetAsInteger : LongInt;
begin
    CheckType(oftInteger);

    Result := LongInt(FDataPtr^);
end;

{-------------------------------------------------------------------------}

procedure TOdbField.SetAsInteger(const Value : LongInt);
begin
    CheckType(oftInteger);

    LongInt(FDataPtr^) := Value;
end;

{--------------------------------------------------------------------------}

{$ifdef __Testing}

procedure TOdbFieldUnitTest.TestOdbField;
var
    StringField  : TOdbField;
    FloatField   : TOdbField;
    IntegerField : TOdbField;

    procedure CheckInvalidFieldTypes;
    var
        Raised : Boolean;
    begin
        Raised := False;
        try
            FloatField.GetStrAddress;
        except
            on E:EOdbInvalidFieldType do
                Raised := True;
        end;
        Check( Raised );

        Raised := False;
        try
            FloatField.AsInteger;
        except
            on E:EOdbInvalidFieldType do
                Raised := True;
        end;
        Check( Raised );

        Raised := False;
        try
            StringField.AsFloat;
        except
            on E:EOdbInvalidFieldType do
                Raised := True;
        end;
        Check( Raised );


        Raised := False;
        try
            IntegerField.AsFloat;
        except
            on E:EOdbInvalidFieldType do
                Raised := True;
        end;
        Check( Raised );

    end;

begin
    StringField  := TOdbField.Create('StringField', oftString, 13);
    FloatField   := TOdbField.Create('FloatField', oftFloat, 0 );
    IntegerField := TOdbField.Create('IntegerField', oftInteger, 0 );
    try
        CheckInvalidFieldTypes;

        StringField.AsString   := 'ValueString';
        IntegerField.AsInteger := 945;
        FloatField.AsFloat     := 34.789;

        Check( StringField.Name  = 'StringField');
        Check( IntegerField.Name = 'IntegerField');
        Check( FloatField.Name   = 'FloatField');

        Check( StringField.Size  = 13 );

        StringField.AsString := OdbMisc.CreateStr('s', 13);
        { test truncation to Size-1 characters }
        Check( StringField.AsString = OdbMisc.CreateStr('s', 12) );

        Check( IntegerField.AsInteger         = 945);
        Check( Round(FloatField.AsFloat*1000) = 34789);

        Check( StringField.FieldType  = oftString);
        Check( IntegerField.FieldType = oftInteger);
        Check( FloatField.FieldType   = oftFloat);

        { It is acceptable to query AsString on all field types }
        Check(IntegerField.AsString = '945');
        Check( Copy(FloatField.AsString, 1, 6)   = '34.789');

        Check( PChar(StringField.GetStrAddress)^ = 's' );

        Check( LongInt(IntegerField.DataPtr^) = 945 );

    finally
        IntegerField.Free;
        FloatField.Free;
        StringField.Free;
    end;
end;

{-------------------------------------------------------------------------}

procedure TOdbFieldUnitTest.TestCompareOdbField;
var
   Field1 : TOdbField;
   Field2 : TOdbField;
begin
    Field1 := TOdbField.Create('Field1', oftString, 21);
    Field2 := TOdbField.Create('Field2', oftInteger, 0);
    try
        Check( CompareOdbField( Field1, Field2 ) = ObjCmp.cmpLess );
        Check( CompareOdbField( Field2, Field1 ) = ObjCmp.cmpGreater );

        Field2.Name := Field1.Name;
        Check( CompareOdbField( Field1, Field2 ) = ObjCmp.cmpEqual );
    finally
        Field1.Free;
        Field2.Free;
    end;
end;

{-------------------------------------------------------------------------}

procedure TOdbFieldUnitTest.TestGetFieldInList;
var
    FieldList : TUnOrderedList;
    Raised    : Boolean;
    Field1    : TOdbField;
begin
    FieldList := TUnOrderedList.Create( TOdbField, CompareOdbField );
    try
        Field1 := TOdbField.Create('Field1', oftFloat, 0);

        FieldList.Insert(Field1);
        FieldList.Insert(TOdbField.Create('Field2', oftInteger, 0));
        FieldList.Insert(TOdbField.Create('Field3', oftString, 11));

        Check( GetFieldInList( 'Field1', FieldList ) = Field1 );

        Raised := False;
        try
            GetFieldInList( 'Field7', FieldList );
        except
            on E:EOdbFieldDoesNotExist do
                Raised := True;
        end;

        Check(Raised);

    finally
        FieldList.Free;
    end;
end;

{-------------------------------------------------------------------------}

procedure TOdbFieldUnitTest.TestFieldsEqual;
var
    Field1  : TOdbField;
    Field2  : TOdbField;
    Raised  : Boolean;
begin
    Field1 := TOdbField.Create('Field1', oftInteger, 0 );
    Field2 := TOdbField.Create('Field2', oftInteger, 0 );
    try
        Field1.AsInteger := 45;
        Field2.AsInteger := 45;
        Check( FieldsEqual(Field1, Field2 ) );

        Field1.AsInteger := 45;
        Field2.AsInteger := 465;
        Check( not FieldsEqual(Field1, Field2 ) );
    finally
        Field1.Free;
        Field2.Free;
    end;

    Field1 := TOdbField.Create('Field1', oftFloat, 0 );
    Field2 := TOdbField.Create('Field2', oftFloat, 0 );
    try
        Field1.AsFloat := 12.45;
        Field2.AsFloat := 12.45;
        Check( FieldsEqual(Field1, Field2 ) );

        Field1.AsFloat := 12.45;
        Field2.AsFloat := 12.465;
        Check( not FieldsEqual(Field1, Field2 ) );
    finally
        Field1.Free;
        Field2.Free;
    end;

    Field1 := TOdbField.Create('Field1', oftString, 34 );
    Field2 := TOdbField.Create('Field2', oftString, 9 );
    try
        Field1.AsString := 'String45';
        Field2.AsString := 'String45';
        Check( FieldsEqual(Field1, Field2 ) );

        Field1.AsString := 'String45';
        Field2.AsString := 'String145';
        Check( not FieldsEqual(Field1, Field2 ) );
    finally
        Field1.Free;
        Field2.Free;
    end;

    Field1 := TOdbField.Create('Field1', oftString, 34 );
    Field2 := TOdbField.Create('Field2', oftInteger, 0 );
    try
        Field1.AsString  := '45';
        Field2.AsInteger := 45;

        Raised := False;
        try
            Check( FieldsEqual(Field1, Field2 ) );
        except
            on E:EOdbFieldsNotComparable do
                Raised := True;
        end;

        Check(Raised);

    finally
        Field1.Free;
        Field2.Free;
    end;
end;

{-------------------------------------------------------------------------}

procedure TOdbFieldUnitTest.TestFieldListHasAutoInc;
var
    FieldList : TUnOrderedList;
begin
    FieldList := TUnOrderedList.Create( TOdbField, CompareOdbField );
    try
        FieldList.Insert(TOdbField.Create('Field1', oftFloat, 0));
        FieldList.Insert(TOdbField.Create('Field2', oftInteger, 0));
        FieldList.Insert(TOdbField.Create('Field3', oftString, 11));
        Check( not FieldListHasAutoInc( FieldList ) );

        FieldList.Insert(TOdbField.Create('Field4', oftAutoInc, 0));
        Check( FieldListHasAutoInc( FieldList ) );

    finally
        FieldList.Free;
    end;
end;

{$endif}

{-------------------------------------------------------------------------}

begin

    RegisterStreamable( TOdbField,
                        [ReadOdbField],
                        [WriteOdbField]
                      );

    RegisterCompareFunction( 'CompareOdbField', CompareOdbField );

end.

